#!/usr/bin/perl -w

print "Testing whether all checkers work as expected ".
	"- this may take a while.\n";

$tmpfile = "/tmp/mckit-test.$$";
$problems = 0;
$report = "";

$spinfound = find_spin();
print "Spin not found - some tests will be skipped.\n" unless $spinfound;

# Exercising the input languages
for $l (qw/apnn cfa senil/) {
	do_check("$l:pep-reach","dekker_non_mutex.$l","mutex.form","YES","Witness");
}
do_check("bpn2:pep-reach","dekker_non_mutex.bpn2","mutex_bpn2.form","YES","Witness");

# Exercising the deadlock checkers
for $c (qw/pep prod smv mcs clp/) {
	do_check("cfa:$c-dl","dekker.cfa","","YES","");
	do_check("cfa:$c-dl","dekker_deadlock.cfa","","NO","Counterexample");
}

# Exercising the reachability checkers
for $c (qw/pep prod mcs/) {
	do_check("cfa:$c-reach","dekker.cfa","mutex.form","NO","");
	do_check("cfa:$c-reach","dekker_non_mutex.cfa","mutex.form","YES","Witness");
}

# Exercising the LTL checkers
for $c (qw/pep prod spin dssz/) {
	do_check("cfa:$c-ltl","dekker.cfa","liveness_1.ltl","NO","Counterexample");
	do_check("cfa:$c-ltl","dekker_fair.cfa","liveness_2.ltl","YES","");
}

# Exercising the CTL checkers
for $c (qw/prod smv dssz/) {
	do_check("cfa:$c-ctl","dekker.cfa","cr1entry.ctl","YES","");
}

if (!$problems) {
	print "No problems found!\n";
} else {
	printf "%d problem%s found!\n\nDetails:\n",
		$problems,($problems > 1)? "s":"";
	print $report;
}
exit 0;

##############################################################################

sub do_check {
	my ($lang,$file,$form,$yn,$cw) = @_;
	my $line = "trying $lang $file $form...";
	$line .= " " x (62-length($line));

	print $line;

	if (!$spinfound && $lang =~ /pep-ltl|spin-ltl/) {
		print " skipped\n";
		return;
	}

	redir_stderr_start();
	system("../bin/check $lang $file $form > $tmpfile");
	redir_stderr_stop();

	open FD,$tmpfile;
	my ($ynsat,$cwsat) = (0,0);
	while (<FD>) {
		$ynsat = 1 if /^Result: $yn\./;
		$cwsat = 1 if /$cw/;
	}
	close FD;
	unlink $tmpfile;

	if (!$ynsat) {
		print " FAILED\n";
		$report .= "Problem with $lang $file $form:\n\t".
				"$yn answer not detected.\n";
		$problems++;
	} elsif ($cw && !$cwsat && $lang !~ /dssz/) {
		print " FAILED\n";
		$report .= "Problem with $lang $file $form:\n\t".
				"$cw not detected.\n";
		$problems++;
	} else {
		printf " passed\n";
	}
}

sub redir_stdout_start {
	open OLDOUT,">&STDOUT";
	open OLDOUT,">&STDOUT";
	open STDOUT,'>/dev/null';
}

sub redir_stderr_start {
	open OLDERR,">&STDERR";
	open OLDERR,">&STDERR";
	open STDERR,'>/dev/null';
}

sub redir_stdout_stop {
	close STDOUT;
	open STDOUT,">&OLDOUT";
}

sub redir_stderr_stop {
	close STDERR;
	open STDERR,">&OLDERR";
}

sub find_spin {
	redir_stdout_start();
	redir_stderr_start();
	my $res = system("spin -V") && system("../tools/spin -V");

	if (defined($ENV{'MCKIT_HOME'})) {
		my $spin = $ENV{'MCKIT_HOME'}."/spin -V";
		$res = $res && system("$spin");
	}
	redir_stdout_stop();
	redir_stderr_stop();

	return !$res;
}
