/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  baltree.h
*
* Part of: PROD's programs araprod, "gphgen", probe, prpp, and strong
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See baltree.c.
*
* Version history:
*     $Log$
***************************************************************************/

/* The AVL type is used to present an AVL tree without side information. */
typedef struct AVL
{
    VOID *element;     /* the element of a vertex */
    int balance;       /* between insertions and deletions, this is
                        * guaranteed to be equal to rh - lh where rh is
                        * the height of the right-hand child subtree and
                        * lh is the height of the left-hand child subtree */
    struct AVL *left;  /* the left-hand child subtree */
    struct AVL *right; /* the right-hand child subtree */
} AVL;

/* The BALTOOLS type is used to collect together routines for copying,
 * deleting and displaying of a element in a tree, and for comparing
 * two elements, and for checking whether a element has a given
 * prefix in some context-dependent sense. */
typedef struct BALTOOLS
{
    CFUNC copy;
    VOF discard;
#ifdef DEBUG
    VOF dump;
#endif
    INTF keyCmp;
    FOOL prefix;
} BALTOOLS;

/* The BALTREE type is used to present an AVL tree with side information. */
typedef struct BALTREE
{
    unsigned long elCount; /* the number of elements */
    AVL *root;
    BALTOOLS *tools;
} BALTREE;

#define BalTrGElCount(balTree) ((balTree)->elCount)
#define BalTrIElCount(balTree) (((balTree)->elCount)++)
#define BalTrSElCount(balTree,_elCount) ((balTree)->elCount = (_elCount))
#define BalTrDElCount(balTree) (((balTree)->elCount)--)
#define BalTrGRoot(balTree) ((balTree)->root)
#define BalTrARoot(balTree) (&((balTree)->root))
#define BalTrSRoot(balTree,_root) ((balTree)->root = (_root))
#define BalTrGTools(balTree) ((balTree)->tools)
#define BalTrSTools(balTree,_tools) ((balTree)->tools = (_tools))
#define BalElCopy(balTree) (BalTrGTools(balTree)->copy)
#define BalElDiscard(balTree) (BalTrGTools(balTree)->discard)
#ifdef DEBUG
#define BalElDump(balTree) (BalTrGTools(balTree)->dump)
#endif
#define BalElKeyCmp(balTree) (BalTrGTools(balTree)->keyCmp)
#define BalElPrefix(balTree) (BalTrGTools(balTree)->prefix)

#define BalTrAlloc() ((BALTREE *) Allocate(sizeof(BALTREE)))
extern TOOLS balTrTools;

DE1(BALTREE *, BalTrCreate,
    BALTOOLS *, tools)
DE2(BALTREE *, BalTrCopy,
    BALTREE *, balTree,
    BOOL, elCopy)
DE2(void, BalTrDestroy,
    BALTREE *, balTree,
    BOOL, elDiscard)
#ifdef DEBUG
DE2(void, BalTrDump,
    BALTREE *, balTree,
    unsigned, depth)
#endif
DE2(VOID **, BalTrApply,
    BALTREE *, balTree,
    FOOL, f)
DE2(VOID **, BalTrBackApply,
    BALTREE *, balTree,
    FOOL, f)
DE3(BALTREE *, BalTrImage,
    BALTREE *, balTree,
    BALTOOLS *, tools,
    CFUNC, strictlyIncreasing)
DE4(VOID *, BalTrInsert,
    BALTREE *, balTree,
    char *, element,
    char *, key,
    BOOL, unique)
DE4(VOID *, BalTrDelete,
    BALTREE *, balTree,
    char *, element,
    char *, key,
    BOOL, elDiscard)
DE2(VOID *, BalTrElement,
    BALTREE *, balTree,
    char *, key)
DE3(VOID *, BalTrGrEqPrefix,
    BALTREE *, balTree,
    char *, prefix,
    BOOL, equalPrefix)
#ifdef SMODELS
DE3(VOID *, BalTrLtEqPrefix,
    BALTREE *, balTree,
    char *, prefix,
    BOOL, equalPrefix)
#endif
DE4(VOID *, BalTrNext,
    BALTREE *, balTree,
    char *, element,
    char *, key,
    char *, prefix)
DE1(VOID *, BalTrFirst,
    BALTREE *, balTree)
DE1(VOID *, BalTrLast,
    BALTREE *, balTree)
#ifdef DEBUG
DE1(void, BalTrChkBal,
    BALTREE *, balTree)
#endif
