/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  dynam.h
*
* Part of: PROD's programs araprod, "gphgen", probe, prpp, strong, and t
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See dynam.c.
*
* Version history:
*     $Log$
***************************************************************************/

typedef VOID *(*CFUNC)();
typedef BOOL (*FOOL)();
typedef int (*INTF)();
typedef unsigned (*UNSF)();
typedef DOMAIN (*DOMF)();
typedef void (*VOF)();

/* The TOOLS type is used to collect together routines for copying,
 * deleting and displaying of an element in a list. */
typedef struct TOOLS
{
    CFUNC _copy;
    VOF _discard;
#ifdef DEBUG
    VOF _dump;
#endif
} TOOLS;

/* The LISTNODE type is used to present a basic unit in a list. */
typedef struct LISTNODE
{
    struct LISTNODE *_next;
    VOID *_node;
} LISTNODE;

/* The LIST type is used to present a list. */
typedef struct LIST
{
    LISTNODE *_first, *_this, *_last; /* cursors pointing to the
                                       * first, "current" and last
                                       * basic unit */
    TOOLS *_tools;
} LIST;

extern TOOLS strTools;

#define LstBegin(list) ((list)->_this = (list)->_first, LstThis(list))
#define LstEmpty(list) (!(list)->_first)
#define LstFirst(list) ((list)->_first ? (list)->_first->_node : NULL)
#define LstGet(list) ((VOID *) ((list)->_this))
#define LstLast(list) ((list)->_last ? (list)->_last->_node : NULL)
#define LstNext(list) \
  ((list)->_this->_next ? (list)->_this->_next->_node : NULL)
#define LstSet(list,pos) ((list)->_this = (LISTNODE *) (pos))
#define LstStart(list) ((list)->_this = NULL)
#define LstStep(list) \
  ((list)->_this = ((list)->_this ? (list)->_this->_next \
    : (list)->_first), LstThis(list))
#define LstThis(list) ((list)->_this ? (list)->_this->_node : NULL)
extern TOOLS lstTools;

#define LSTFOR(list,node,type) \
  for ((node) = (type *) LstBegin(list); \
       (node); \
       (node) = (type *) LstStep(list))
#define LSTNEST(list,node,type) \
  for (LstStart(list), LstStore(list), (node) = (type *) LstStep(list); \
       (node); \
       LstStore(list), (node) = (type *) LstStep(list))
#define LSTNUM(list,node,type,num) \
  for ((node) = (type *) LstBegin(list), (num) = 0; \
       (node); \
       (node) = (type *) LstStep(list), (num)++)
#define LSTROLL(list,node,type,num) \
  for (LstStart(list), LstStore(list), \
       (node) = (type *) LstStep(list), (num) = 0; \
       (node); \
       LstStore(list), (node) = (type *) LstStep(list), (num)++)
#define LSTUNNEST(list,node,type) \
  for ((void) LstRestore(list), (node) = (type *) LstThis(list); \
       (node); \
       (void) LstRestore(list), (node) = (type *) LstThis(list))
#define LSTUNROLL(list,node,type,num) \
  for ((num) = LstSize(list) - 1, (void) LstRestore(list), \
       (node) = (type *) LstThis(list); \
       (node); \
       (void) LstRestore(list), (node) = (type *) LstThis(list), (num)--)

DE0(void, DynamInit)
DE1(void, DumpIndent,
    unsigned, depth)
DE1(char *, StrCp,
    char *, string)
#ifdef DEBUG
DE2(void, StrDump,
    char *, string,
    unsigned, depth)
#endif
DE2(LIST *, LstCatenate,
    LIST *, list1,
    LIST *, list2)
DE2(LIST *, LstCopy,
    LIST *, list,
    BOOL, trueCopy)
DE1(LIST *, LstCreate,
    TOOLS *, tools)
DE3(VOID *, LstDelete,
    LIST *, list,
    char *, node,
    BOOL, discard)
DE2(VOID *, LstDelFirst,
    LIST *, list,
    BOOL, discard)
DE2(VOID *, LstDelLast,
    LIST *, list,
    BOOL, discard)
DE2(VOID *, LstDelNext,
    LIST *, list,
    BOOL, discard)
DE2(void, LstDestroy,
    LIST *, list,
    BOOL, discard)
#ifdef DEBUG
DE2(void, LstDump,
    LIST *, list,
    unsigned, depth)
#endif
DE2(VOID *, LstInsFirst,
    LIST *, list,
    char *, node)
DE2(VOID *, LstInsLast,
    LIST *, list,
    char *, node)
DE2(VOID *, LstInsNext,
    LIST *, list,
    char *, node)
DE2(VOID *, LstNth,
    LIST *, list,
    unsigned, n)
DE1(VOID *, LstRestore,
    LIST *, list)
DE1(unsigned, LstSize,
    LIST *, list)
DE1(void, LstStore,
    LIST *, list)
DE2(unsigned, LstWhich,
    LIST *, list,
    char *, node)
DE1(LIST *, LstCp,
    LIST *, list)
DE1(void, LstDc,
    LIST *, list)
