/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  gphutil.h
*
* Part of: PROD's programs araprod, "gphgen", probe, prpp, and strong
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See gphutil.c.
*
* Version history:
*     $Log$
***************************************************************************/

/* The BINARYFILE type hides the implementation of binary files. */
#ifdef UNIX
typedef int BINARYFILE;
#define NotBinaryFile(file) ((file) < 0)
#else
typedef FILE *BINARYFILE;
#define NotBinaryFile(file) (!(file))
#endif

/* The MMFILE type is used for replacing a binary file by a table in
 * the memory. */
typedef struct MMFILE
{
    char *name;
    BINARYFILE bFile;
    char *start;
    unsigned long len;
    unsigned long actLen;
    unsigned long offset;
} MMFILE;

#ifndef UNIX
#endif

#define LngAlloc() ((unsigned long *) Allocate(sizeof(unsigned long)))
extern TOOLS lngTools;
extern BALTOOLS lngBalTools;

#define DomAlloc() ((DOMAIN *) Allocate(sizeof(DOMAIN)))
extern TOOLS domTools;
extern BALTOOLS domBalTools;

extern TOOLS dataTools;
extern BALTOOLS dataBalTools;
extern BALTOOLS strBalTools;

/* The REF type is used when there is an unordered table and we want to
 * know the position of a key. */
typedef struct REF
{
    char *key;
    unsigned long slot;
} REF;

#define RefGKey(ref) ((ref)->key)
#define RefGSlot(ref) ((ref)->slot)
#define RefAlloc() ((REF *) Allocate(sizeof(REF)))
#define RefDiscard(ref) Deallocate((char *) (ref))
extern BALTOOLS refBalTools;

/* The DREF type is used when there is an unordered table and we want
 * to know the position of a key. DREF differs from the REF type in the
 * type of the key and in the intended usage. When an element of DREF
 * is deleted, the key is deleted, whereas the deletion of an element
 * of REF does not delete the key. */
typedef struct DREF
{
    DATATYPE *key;
    unsigned long slot;
} DREF;

#define DrefGKey(dref) ((dref)->key)
#define DrefGSlot(dref) ((dref)->slot)
#define DrefAlloc() ((DREF *) Allocate(sizeof(DREF)))
extern BALTOOLS drefBalTools;


#define LEAFBITS 13
#define LEAFSIZE (((unsigned) 1) << LEAFBITS)
#define LEAFMASK (LEAFSIZE - 1)

/* The BITSET type is used when there is a fixed set and we want to present
 * some of its subsets efficiently with respect to space and search time. */
typedef struct BITSET
{
    unsigned long count;
    PTRBL *set;
} BITSET;

#define BstAlloc() ((BITSET *) Allocate(sizeof(BITSET)))
#define BstGCount(bitset) ((bitset)->count)
#define	BstGNAtoms(bitset) \
(1L + (BstGCount(bitset) - 1L) / (BITSPERBYTE * sizeof(DATATYPE)))
#define BstGNLeaves(bitset) \
(((BstGNAtoms(bitset) - 1L) >> LEAFBITS) + 1L)
#define BstGLastSize(bitset) \
((unsigned) (((BstGNAtoms(bitset) - 1L) & LEAFMASK) + 1L))
#define BstACount(bitset) (&((bitset)->count))
#define BstSCount(bitset,_count) ((bitset)->count = (_count))
#define BstGSet(bitset) ((bitset)->set)
#define BstASet(bitset) (&((bitset)->set))
#define BstSSet(bitset,_set) ((bitset)->set = (_set))

/* The MON type is used in on-the-fly verification. The elements of MON
 * are called monitor states. A monitor state consists of a state number
 * and a flag indicating what kind of a monitor state is in question. */
typedef struct MON
{
    DOMAIN state;
    unsigned flag;
} MON;

#define MonAlloc() ((MON *) Allocate(sizeof(MON)))
#define MonGState(mon) ((mon)->state)
#define MonAState(mon) (&((mon)->state))
#define MonSState(mon,_state) ((mon)->state = (_state))
#define MonGFlag(mon) ((mon)->flag)
#define MonAFlag(mon) (&((mon)->flag))
#define MonSFlag(mon,_flag) ((mon)->flag = (_flag))
#define MonReject(mon) ((mon)->flag & 1)
#define MonSReject(mon) ((mon)->flag |= 1)
#define MonDeadlock(mon) ((mon)->flag & 2)
#define MonSDeadlock(mon) ((mon)->flag |= 2)
#define MonLivelock(mon) ((mon)->flag & 4)
#define MonSLivelock(mon) ((mon)->flag |= 4)
#define MonInfinite(mon) ((mon)->flag & 8)
#define MonSInfinite(mon) ((mon)->flag |= 8)
extern BALTOOLS monBalTools;
extern BALTREE *monitors;

/* The elements of STEP are called steps. A step consists of a node
 * and one of its immediate successor arrows in a reachability graph
 * of a net. */
typedef struct STEP
{
    unsigned long world;
    unsigned long awAddr;
} STEP;

#define StpAlloc() ((STEP *) Allocate(sizeof(STEP)))
#define StpGWorld(step) ((step)->world)
#define StpAWorld(step) (&((step)->world))
#define StpSWorld(step,_world) ((step)->world = (_world))
#define StpGAwAddr(step) ((step)->awAddr)
#define StpAAwAddr(step) (&((step)->awAddr))
#define StpSAwAddr(step,_awaddr) ((step)->awAddr = (_awaddr))
extern TOOLS stpTools;

/* An element of PATH corresponds to a path in a reachability graph of
 * a net. It consists of a list of consecutive steps and an end node. */
typedef struct PATH
{
    LIST *steps;
    unsigned long end;
} PATH;

#define PthEnd(path) ((path)->end)
#define PthAEnd(path) (&((path)->end))
#define PthSEnd(path,_end) ((path)->end = (_end))
#define PthGSteps(path) ((path)->steps)
#define PthASteps(path) (&((path)->steps))
#define PthSSteps(path,_steps) ((path)->steps = (_steps))
#define PthLast(path) (((STEP *) LstFirst((path)->steps)))
extern TOOLS pthTools;
extern BALTOOLS pthBalTools;
typedef PATH *(*PTHF)();

DE4(MMFILE *, MapToMem,
    char *, name,
    BINARYFILE, bFile,
    unsigned long, len,
    unsigned long, actLen)
DE1(void, MapFlush,
    MMFILE *, mmFile)
DE2(void, Unmap,
    MMFILE *, mmFile,
    BOOL, wr)
DE1(BINARYFILE, CreateUpdate,
    char *, name)
DE1(BINARYFILE, CreateWrite,
    char *, name)
DE2(BINARYFILE, OpenRead,
    char *, name,
    BOOL, strict)
DE1(BINARYFILE, OpenUpdate,
    char *, name)
DE1(void, OwnClose,
    BINARYFILE, file)
#ifndef UNIX
DE1(void, OwnFlush,
    BINARYFILE, file)
#endif
DE2(void, SeekAbs,
    BINARYFILE, file,
    unsigned long, pos)
DE1(unsigned long, SeekEnd,
    BINARYFILE, file)
DE3(void, ReadBytes,
    char *, buf,
    unsigned long, cnt,
    BINARYFILE, file)
DE3(void, WriteBytes,
    char *, buf,
    unsigned long, cnt,
    BINARYFILE, file)
DE2(void, MapSkAbs,
    MMFILE *, mmFile,
    unsigned long, pos)
DE3(void, MapReBytes,
    char *, buf,
    unsigned long, cnt,
    MMFILE *, mmFile)
DE3(void, MapWrBytes,
    char *, buf,
    unsigned long, cnt,
    MMFILE **, mmFile)
#define MltLng(a,b) \
    ((unsigned long) (((unsigned long) (a)) * (unsigned long) (b)))
#define ReadType(type,buf,cnt,file) \
    ReadBytes((char *) (buf), MltLng(sizeof(type), cnt), (BINARYFILE) (file))
#define NReadType(type,buf,cnt,file) \
    (ReadType(type, buf, cnt, file), *(((type *) (buf)) + (cnt)) = (type) 0)
#define WriteType(type,buf,cnt,file) \
    WriteBytes((char *) (buf), MltLng(sizeof(type), cnt), (BINARYFILE) (file))
#define MapReType(type,buf,cnt,mmFile) \
    MapReBytes((char *) (buf), MltLng(sizeof(type), cnt), (MMFILE *) (mmFile))
#define NMapReType(type,buf,cnt,file) \
    (MapReType(type, buf, cnt, file), *(((type *) (buf)) + (cnt)) = (type) 0)
#define MapWrType(type,buf,cnt,mmFile) \
    MapWrBytes((char *) (buf), MltLng(sizeof(type), cnt), (MMFILE **) (mmFile))
#define ReadLng(buf,cnt,file) ReadType(unsigned long, buf, cnt, file)
#define WriteLng(buf,cnt,file) WriteType(unsigned long, buf, cnt, file)
#define ReadUns(buf,cnt,file) ReadType(unsigned, buf, cnt, file)
#define WriteUns(buf,cnt,file) WriteType(unsigned, buf, cnt, file)
#define ReadData(buf,cnt,file) NReadType(DATATYPE, buf, cnt, file)
#define WriteData(buf,cnt,file) WriteType(DATATYPE, buf, cnt, file)
#define ReadChr(buf,cnt,file) NReadType(char, buf, cnt, file)
#define WriteChr(buf,cnt,file) WriteType(char, buf, cnt, file)
#define MapReLng(buf,cnt,file) MapReType(unsigned long, buf, cnt, file)
#define MapWrLng(buf,cnt,file) MapWrType(unsigned long, buf, cnt, file)
#define MapReUns(buf,cnt,file) MapReType(unsigned, buf, cnt, file)
#define MapWrUns(buf,cnt,file) MapWrType(unsigned, buf, cnt, file)
#define MapReData(buf,cnt,file) NMapReType(DATATYPE, buf, cnt, file)
#define MapWrData(buf,cnt,file) MapWrType(DATATYPE, buf, cnt, file)
#define MapReChr(buf,cnt,file) NMapReType(char, buf, cnt, file)
#define MapWrChr(buf,cnt,file) MapWrType(char, buf, cnt, file)
DE1(unsigned long *, LngCons,
    unsigned long, value)
DE1(unsigned long *, LngCopy,
    unsigned long *, lng)
#ifdef DEBUG
DE2(void, LngDump,
    unsigned long *, lng,
    unsigned, depth)
#endif
DE2(int, LngCompare,
    unsigned long *, lng1,
    unsigned long *, lng2)
DE2(BOOL, LngPrefix,
    unsigned long *, lng,
    unsigned long *, prefix)
DE1(DOMAIN *, DomCons,
    DOMAIN, value)
DE1(DOMAIN *, DomCopy,
    DOMAIN *, dom)
#ifdef DEBUG
DE2(void, DomDump,
    DOMAIN *, dom,
    unsigned, depth)
#endif
DE2(int, DomCompare,
    DOMAIN *, dom1,
    DOMAIN *, dom2)
DE2(BOOL, DomPrefix,
    DOMAIN *, dom,
    DOMAIN *, prefix)
DE1(DATATYPE *, DataCopy,
    DATATYPE *, t)
#ifdef DEBUG
DE2(void, DataDump,
    DATATYPE *, t,
    unsigned, depth)
#endif
DE1(unsigned, DataLen,
    DATATYPE *, t)
DE2(int, DataCompare,
    DATATYPE *, s,
    DATATYPE *, t)
DE2(BOOL, DataPrefix,
    DATATYPE *, s,
    DATATYPE *, p)
DE2(BOOL, StrPrefix,
    char *, str,
    char *, prefix)
DE2(REF *, RefCons,
    char *, key,
    unsigned long, slot)
DE1(REF *, RefCopy,
    REF *, ref)
#ifdef DEBUG
DE2(void, RefDump,
    REF *, ref,
    unsigned, depth)
#endif
DE2(int, RefKeyCmp,
    REF *, ref,
    char *, key)
DE2(BOOL, RefPrefix,
    REF *, ref,
    char *, prefix)
DE2(DREF *, DrefCons,
    DATATYPE *, key,
    unsigned long, slot)
DE1(DREF *, DrefCopy,
    DREF *, dref)
DE1(void, DrefDiscard,
    DREF *, dref)
#ifdef DEBUG
DE2(void, DrefDump,
    DREF *, dref,
    unsigned, depth)
#endif
DE2(int, DrefKeyCmp,
    DREF *, dref,
    DATATYPE *, key)
DE2(BOOL, DrefPrefix,
    DREF *, dref,
    DATATYPE *, prefix)
DE2(int, StrLstCompare,
    LIST *, list1,
    LIST *, list2)
DE2(BOOL, StrLstPrefix,
    LIST *, list,
    LIST *, prefix)
DE1(BITSET *, BstCons,
    unsigned long, count)
DE1(void, BstDiscard,
    BITSET *, bitset)
#ifdef DEBUG
DE2(void, BstDump,
    BITSET *, bitset,
    unsigned, depth)
#endif
DE1(void, BstClear,
    BITSET *, bitset)
DE2(BOOL, BstIn,
    BITSET *, bitset,
    unsigned long, number)
DE2(void, BstInsert,
    BITSET *, bitset,
    unsigned long, number)
DE2(void, BstDelete,
    BITSET *, bitset,
    unsigned long, number)
DE1(MON *, MonCons,
    DOMAIN, state)
#ifdef DEBUG
DE2(void, MonDump,
    MON *, mon,
    unsigned, depth)
#endif
DE2(int, MonKeyCmp,
    MON *, mon,
    DOMAIN *, key)
DE2(STEP *, StpCons,
    unsigned long, world,
    unsigned long, awAddr)
DE1(STEP *, StpCopy,
    STEP *, step)
#ifdef DEBUG
DE2(void, StpDump,
    STEP *, step,
    unsigned, depth)
#endif
DE1(PATH *, PthCreate,
    unsigned long, world)
DE1(PATH *, PthCopy,
    PATH *, path)
DE1(void, PthDracsid,
    PATH *, path)
#define PthDiscard(path) (PthDracsid(path), (path) = (PATH *) NULL)
#ifdef DEBUG
DE2(void, PthDump,
    PATH *, path,
    unsigned, depth)
#endif
DE3(void, PthInsFirst,
    PATH *, path,
    unsigned long, world,
    unsigned long, awAddr)
DE3(void, PthInsLast,
    PATH *, path,
    unsigned long, awAddr,
    unsigned long, world)
DE1(void, PthDelLast,
    PATH *, path)
DE2(PATH *, PthCatenate,
    PATH *, path1,
    PATH *, path2)
DE1(BOOL, PthLoop,
    PATH *, path)
DE2(int, PthCompare,
    PATH *, path1,
    PATH *, path2)
DE2(BOOL, PthSuffix,
    PATH *, path,
    PATH *, suffix)
