/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  graph.h
*
* Part of: PROD's programs araprod, "gphgen", probe, and strong
*
* Authors: Keijo Heljanko, Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See graph.c.
*
* Version history:
*     $Log$
***************************************************************************/

#define HASH PTRBL       /* The HASH type is needed for a hash table
                          * in the generation of a reachability
                          * graph of a net. The hash table is used
                          * for checking whether the graph has a node
                          * corresponding to a computed marking of
                          * the net. */
#define GPH_GENERATE 0   /* This is used for indicating that
                          * a reachability graph is under construction. */
#define GPH_COMPS 1      /* This is used for indicating that strongly
                          * connected components of a reachability graph
                          * are being computed. */
#define GPH_PROBE 2      /* This is used for indicating that a reachability
                          * graph is being inspected by the probe program. */
#define GPH_PREPARE_CTL 3

/* The GRAPH type is needed for a reachability graph of a net. */
typedef struct GRAPH
{
    char *name;   /* name of the graph */
    int reason;   /* this is one of GPH_GENERATE, GPH_COMPS and GPH_PROBE,
                   * indicating how the graph is used */
    BINARYFILE nodes;  /* contains node records and graph globals */
    BINARYFILE arrows; /* contains arrow records */
    BINARYFILE addrs;  /* contains node address table (address of nth node is
                        * in nth position of the table) and table for
                        * strongly connected components (address of 1st node
                        * of nth component is in nth position of the table) */
    BINARYFILE stack;  /* contains a stack of nodes and sleep sets */
    unsigned long *glob; /* points to graph globals table (in memory) */
    HASH *hash;          /* points to hash table (in memory) */
    BDD *bdd;            /* points to a BDD */
    BALTREE *balTree;    /* points to an AVL tree */
    MMFILE *mmnodes;   /* a table corresponding to the node file */
    MMFILE *mmarrows;  /* a table corresponding to the arrow file */
    MMFILE *mmaddrs;   /* a table corresponding to the addrs file */
    MMFILE *mmstack;   /* a table corresponding to the stack file */
    unsigned long latest_node; /* address of the most resecently asked node */
    unsigned long *latest_rec; /* points to the record (in memory) of that
                                * node (may be used for other purposes when
                                * the graph is constructed) */
    DATATYPE *latest_contents; /* points to the data (in memory) of that node
			        * (alternatively, to the latest arrow data
			        * asked by probe) (may be used for other
                                * purposes when the graph is constructed) */
    unsigned long *arrow_rec; /* points to an arrow record in memory */
    unsigned long *sgc_rec; /* points to a strongly connected component
			     * record in memory */
    LIST *vis_stack;   /* used when computing strongly connected components */
    LIST *list;        /* used when computing strongly connected components */
    unsigned long now; /* used when computing strongly connected components */
    unsigned long m;   /* used when computing strongly connected components */
    unsigned long *p;  /* used when computing strongly connected components */
    unsigned long *q;  /* used when computing strongly connected components */
    unsigned next_lower_pos;   /* node record field position */
    unsigned check_pos;        /* node record field position */
    unsigned sgc_val_pos;      /* node record field position */
    unsigned predecessors_pos; /* node record field position */
    unsigned left_child_pos;   /* node record field position */
    unsigned right_child_pos;  /* node record field position */
    unsigned datalen_pos;      /* node record field position */
    unsigned num_node_fields;  /* node record field count */
    unsigned aw_datalen_pos;   /* arrow record field position */
    unsigned num_arrow_fields; /* arrow record field count */
} GRAPH;


extern GRAPH *_graph[2];
extern unsigned long _ndAddr;
extern unsigned long _ndFirst;
extern unsigned long _awCount;
extern unsigned long _awAddr;
extern unsigned long _awFirst;
#define theGraph _graph[0]
#define the2Graph _graph[1]
#define theNdAddr _ndAddr
#define theNdFirst _ndFirst
#define theAwCount _awCount
#define theAwAddr _awAddr
#define theAwFirst _awFirst

extern LIST *_rootList;
#define RootList() _rootList
extern LIST *_actList;
#define ActList() _actList
extern PTRBL *_arpFoldings;
#define ArpFoldings() _arpFoldings
extern unsigned _iCode;
#define ICode() _iCode
extern char *_iName;
#define IName() _iName
extern BOOL _araProd;
#define AraProd() _araProd
#define AraSetProd() (_araProd = TRUE)
#define AraResetProd() (_araProd = FALSE)

DE1(unsigned long, PrimeBelow,
    unsigned long, b)
DE3(unsigned long, Hash,
    DATATYPE *, s,
    unsigned long, prime,
    unsigned long, coeff)
DE5(GRAPH *, GphCreate,
    char *, name,
    unsigned, hash_bits,
    unsigned, bufsize,
    UNS32, seed,
    unsigned long, createFlag)
DE5(unsigned long, GphGet,
    GRAPH *, graph,
    DATATYPE **, buffer,
    unsigned *, size,
    BALTREE **, sleep1,
    BALTREE **, sleep2)
DE4(unsigned long, GphPutNode,
    GRAPH *, graph,
    DATATYPE *, node,
    unsigned, len,
    unsigned long *, back)
DE3(unsigned long, GphFndNode,
    GRAPH *, graph,
    DATATYPE *, node,
    unsigned long *, back)
DE3(void, GphErase,
    GRAPH *, graph,
    unsigned long, start_addr,
    DATATYPE *, buffer)
DE9(unsigned long, GphPutArrow,
    GRAPH *, graph,
    DATATYPE *, arrow,
    unsigned, len,
    unsigned long, ordinal,
    unsigned long, source_node_addr,
    unsigned long, target_node_addr,
    unsigned long, aw_next_addr,
    unsigned long, aw_back_addr,
    BOOL, invisible)
DE5(void, GphUpdate,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long, arrow_count,
    unsigned long, first_arrow_addr,
    BALTREE *, sleepSet)
DE2(void, GphAddTermNode,
    GRAPH *, graph,
    unsigned long, node_addr)
DE3(void, GphPush,
    GRAPH *, graph,
    unsigned long, node_addr,
    BALTREE *, sleepSet)
DE1(void, GphIncGenCount,
    GRAPH *, graph)
DE1(unsigned long, GphGenCount,
    GRAPH *, graph)
DE3(GRAPH *, GphOpen,
    char *, name,
    int, reason,
    BOOL, rotSuff)
DE1(void, GphFlush,
    GRAPH *, graph)
DE2(void, GphMapToMem,
    GRAPH *, graph,
    unsigned long, len)
DE1(void, GphClose,
    GRAPH *, graph)
DE1(unsigned long, GphStart,
    GRAPH *, graph)
DE2(unsigned long, GphCompNumber,
    GRAPH *, graph,
    unsigned long, node_addr)
DE2(unsigned long, GphOrdinal,
    GRAPH *, graph,
    unsigned long, node_addr)
DE2(unsigned long, GphId,
    GRAPH *, graph,
    unsigned long, ordinal)
DE1(unsigned long, GphNdCount,
    GRAPH *, graph)
DE1(unsigned long, GphGlAwCount,
    GRAPH *, graph)
DE2(unsigned long, GphAwCount,
    GRAPH *, graph,
    unsigned long, node_addr)
DE3(BOOL, GphFresh,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long, mask)
DE3(void, GphConsume,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long, mask)
DE4(void, GphFlFirst,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long *, awaddr,
    unsigned long *, succ)
DE3(void, GphFlNext,
    GRAPH *, graph,
    unsigned long *, awaddr,
    unsigned long *, succ)
DE4(void, GphPrLast,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long *, awaddr,
    unsigned long *, pred)
DE3(void, GphPrPrec,
    GRAPH *, graph,
    unsigned long *, awaddr,
    unsigned long *, pred)
DE2(DATATYPE *, GphNode,
    GRAPH *, graph,
    unsigned long, node_addr)
DE2(DATATYPE *, GphArrow,
    GRAPH *, graph,
    unsigned long, awaddr)
DE6(void, GphChLastArrow,
    GRAPH *, graph,
    DATATYPE *, arrow,
    unsigned, len,
    unsigned long, target,
    unsigned long *, back,
    BOOL, icold)
DE2(unsigned long, GphAwOrdinal,
    GRAPH *, graph,
    unsigned long, awaddr)
DE2(BOOL, GphAwInvisible,
    GRAPH *, graph,
    unsigned long, awaddr)
DE2(BOOL, GphAwHot,
    GRAPH *, graph,
    unsigned long, awaddr)
DE2(void, GphStatistics,
    GRAPH *, graph,
    FILE *, file)
DE1(unsigned long, GphSgCount,
    GRAPH *, graph)
DE1(void, GphSgSearch,
    GRAPH *, graph)
DE2(BALTREE *, GphSgGet,
    GRAPH *, graph,
    unsigned long, component)
DE2(unsigned long, GphSgFlCount,
    GRAPH *, graph,
    unsigned long, component)
DE2(BALTREE *, GphSgSucc,
    GRAPH *, graph,
    unsigned long, component)
DE1(BALTREE *, GphTNGet,
    GRAPH *, graph)
DE1(unsigned long, GphTmSgCount,
    GRAPH *, graph)
DE1(BALTREE *, GphTmSgGet,
    GRAPH *, graph)
DE1(unsigned, GphBufSize,
    GRAPH *, graph)
DE1(unsigned, GphBits,
    GRAPH *, graph)
DE1(UNS32, GphSeed,
    GRAPH *, graph)
DE2(void, GphSetSeed,
    GRAPH *, graph,
    UNS32, seed)
DE1(BOOL, GphUnfold,
    GRAPH *, graph)
DE1(BOOL, GphSymm,
    GRAPH *, graph)
DE1(BOOL, GphIMkSymm,
    GRAPH *, graph)
DE1(BOOL, GphStub,
    GRAPH *, graph)
DE1(BOOL, GphDelAlg,
    GRAPH *, graph)
DE1(BOOL, GphCFFD,
    GRAPH *, graph)
DE1(BOOL, GphDefVis,
    GRAPH *, graph)
DE1(BOOL, GphSleep,
    GRAPH *, graph)
DE1(BOOL, GphDfs,
    GRAPH *, graph)
DE1(BOOL, GphFreeExpansion,
    GRAPH *, graph)
DE1(BOOL, GphErr,
    GRAPH *, graph)
DE1(void, LTSOutput,
    GRAPH *, graph)
DE3(void, ExOnTheFly,
    GRAPH *, graph,
    PATH *, path,
    BOOL, loop)
DE9(PATH *, GphShortest,
    GRAPH *, graph,
    unsigned long *, present,
    unsigned long *, head,
    unsigned long *, tail,
    PATH **, path,
    BOOL *, cont,
    VOID *, formula,
    FOOL, along,
    PTHF, admit)

/* Graph globals are kept at the beginning of the node file.
 * They are of type unsigned long. The following before
 * NUM_GRAPH_GLOBALS are the positions of the graph globals
 * in the graph global record. */
#define FLAG_POS ((unsigned) 0)
#define NODE_COUNT_POS ((unsigned) 1)
#define ARROW_COUNT_POS ((unsigned) 2)
#define SGC_COUNT_POS ((unsigned) 3)
#define TERMCOMP_COUNT_POS ((unsigned) 4)
#define LATEST_TERMCOMP_POS ((unsigned) 5)
#define TERMNODE_COUNT_POS ((unsigned) 6)
#define LATEST_TERMNODE_POS ((unsigned) 7)
#define GEN_COUNT_POS ((unsigned) 8)
#define EOF_NODES_POS ((unsigned) 9)
#define EOF_ARROWS_POS ((unsigned) 10)
#define FIRST_UNHANDLED_POS ((unsigned) 11)
#define HASH_BITS_POS ((unsigned) 12)
#define HASH_SIZE_POS ((unsigned) 13)
#define HASH_COEFF_POS ((unsigned) 14)
#define BUF_SIZE_POS ((unsigned) 15)
#define MAX_DATALEN_POS ((unsigned) 16)
#define SEED_POS ((unsigned) 17)
#define TOPMOST_POS ((unsigned) 18)
#define EOF_STACK_POS ((unsigned) 19)
#define LAST_ARROW_POS ((unsigned) 20)

#define NUM_GRAPH_GLOBALS ((unsigned) 21)

#define _rotsuff(graph) ((graph)->glob[FLAG_POS] & 1L)
#define _srotsuff(graph) ((graph)->glob[FLAG_POS] |= 1L)
#define _drotsuff(graph) ((graph)->glob[FLAG_POS] &= ~1L)
#define _unfold(graph) ((graph)->glob[FLAG_POS] & 2L)
#define _sunfold(graph) ((graph)->glob[FLAG_POS] |= 2L)
#define _dunfold(graph) ((graph)->glob[FLAG_POS] &= ~2L)
#define _symm(graph) ((graph)->glob[FLAG_POS] & 4L)
#define _ssymm(graph) ((graph)->glob[FLAG_POS] |= 4L)
#define _dsymm(graph) ((graph)->glob[FLAG_POS] &= ~4L)
#define _stub(graph) ((graph)->glob[FLAG_POS] & 8L)
#define _sstub(graph) ((graph)->glob[FLAG_POS] |= 8L)
#define _dstub(graph) ((graph)->glob[FLAG_POS] &= ~8L)
#define _delalg(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 16)
#define _sdelalg(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 16)
#define _ddelalg(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 16))
#define _bdd(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 32)
#define _sbdd(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 32)
#define _dbdd(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 32))
#define _dfs(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 64)
#define _sdfs(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 64)
#define _ddfs(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 64))
#define _frexpansion(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 128)
#define _sfrexpansion(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 128)
#define _dfrexpansion(graph) \
((graph)->glob[FLAG_POS] &= ~((unsigned long) 128))
#define _imksymm(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 256)
#define _simksymm(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 256)
#define _dimksymm(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 256))
#define _cffd(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 512)
#define _scffd(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 512)
#define _dcffd(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 512))
#define _defvis(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 1024)
#define _sdefvis(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 1024)
#define _ddefvis(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 1024))
#define _s_l_e_e_p(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 2048)
#define _s_s_l_e_e_p(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 2048)
#define _d_s_l_e_e_p(graph) \
    ((graph)->glob[FLAG_POS] &= ~((unsigned long) 2048))
#define _gerr(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 4096)
#define _serr(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 4096)
#define _derr(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 4096))
#define _inftr(graph) ((graph)->glob[FLAG_POS] & (unsigned long) 8192)
#define _sinftr(graph) ((graph)->glob[FLAG_POS] |= (unsigned long) 8192)
#define _dinftr(graph) ((graph)->glob[FLAG_POS] &= ~((unsigned long) 8192))

/* Positions of node record fields. Fields are of type unsigned long.
 * See also the definition of the GRAPH type in graph.h. */
#define ORDINAL_POS ((unsigned) 0)
#define SUCCESSOR_COUNT_POS ((unsigned) 1)
#define FOLLOWERS_POS ((unsigned) 2)

/* Positions of arrow record fields. Fields are of type unsigned long.
* See also the definition of the GRAPH type in graph.h. */
#define AW_ORDINAL_POS ((unsigned) 0)
#define SOURCE_POS ((unsigned) 1)
#define AW_BACK_POS ((unsigned) 2)
#define TARGET_POS ((unsigned) 3)
#define AW_NEXT_POS ((unsigned) 4)

/* Positions of strongly connected component record fields. Fields are of
 * type unsigned long. */
#define FIRST_NODE_POS ((unsigned) 0)
#define FIRST_SGC_SUCC_POS ((unsigned) 1)
#define SGC_SUCC_COUNT_POS ((unsigned) 2)

#define NUM_SGC_FIELDS ((unsigned) 3)

#define NdsSkAbs(graph,pos) \
if ((graph)->mmnodes) MapSkAbs((graph)->mmnodes, pos); \
else SeekAbs((graph)->nodes, pos)

#define AwsSkAbs(graph,pos) \
if ((graph)->mmarrows) MapSkAbs((graph)->mmarrows, pos); \
else SeekAbs((graph)->arrows, pos)

#define AdrSkAbs(graph,pos) \
if ((graph)->mmaddrs) MapSkAbs((graph)->mmaddrs, pos); \
else SeekAbs((graph)->addrs, pos)

#define StkSkAbs(graph,pos) \
if ((graph)->mmstack) MapSkAbs((graph)->mmstack, pos); \
else SeekAbs((graph)->stack, pos)

#define NdsReLng(buf,cnt,graph) \
if ((graph)->mmnodes) MapReLng(buf, cnt, (graph)->mmnodes); \
else ReadLng(buf, cnt, (graph)->nodes)

#define AwsReLng(buf,cnt,graph) \
if ((graph)->mmarrows) MapReLng(buf, cnt, (graph)->mmarrows); \
else ReadLng(buf, cnt, (graph)->arrows)

#define AdrReLng(buf,cnt,graph) \
if ((graph)->mmaddrs) MapReLng(buf, cnt, (graph)->mmaddrs); \
else ReadLng(buf, cnt, (graph)->addrs)

#define StkReLng(buf,cnt,graph) \
if ((graph)->mmstack) MapReLng(buf, cnt, (graph)->mmstack); \
else ReadLng(buf, cnt, (graph)->stack)

#define NdsWrLng(buf,cnt,graph) \
if ((graph)->mmnodes) MapWrLng(buf, cnt, &((graph)->mmnodes)); \
else WriteLng(buf, cnt, (graph)->nodes)

#define AwsWrLng(buf,cnt,graph) \
if ((graph)->mmarrows) MapWrLng(buf, cnt, &((graph)->mmarrows)); \
else WriteLng(buf, cnt, (graph)->arrows)

#define AdrWrLng(buf,cnt,graph) \
if ((graph)->mmaddrs) MapWrLng(buf, cnt, &((graph)->mmaddrs)); \
else WriteLng(buf, cnt, (graph)->addrs)

#define StkWrLng(buf,cnt,graph) \
if ((graph)->mmstack) MapWrLng(buf, cnt, &((graph)->mmstack)); \
else WriteLng(buf, cnt, (graph)->stack)

#define NdsReUns(buf,cnt,graph) \
if ((graph)->mmnodes) MapReUns(buf, cnt, (graph)->mmnodes); \
else ReadUns(buf, cnt, (graph)->nodes)

#define AwsReUns(buf,cnt,graph) \
if ((graph)->mmarrows) MapReUns(buf, cnt, (graph)->mmarrows); \
else ReadUns(buf, cnt, (graph)->arrows)

#define AdrReUns(buf,cnt,graph) \
if ((graph)->mmaddrs) MapReUns(buf, cnt, (graph)->mmaddrs); \
else ReadUns(buf, cnt, (graph)->addrs)

#define StkReUns(buf,cnt,graph) \
if ((graph)->mmstack) MapReUns(buf, cnt, (graph)->mmstack); \
else ReadUns(buf, cnt, (graph)->stack)

#define NdsWrUns(buf,cnt,graph) \
if ((graph)->mmnodes) MapWrUns(buf, cnt, &((graph)->mmnodes)); \
else WriteUns(buf, cnt, (graph)->nodes)

#define AwsWrUns(buf,cnt,graph) \
if ((graph)->mmarrows) MapWrUns(buf, cnt, &((graph)->mmarrows)); \
else WriteUns(buf, cnt, (graph)->arrows)

#define AdrWrUns(buf,cnt,graph) \
if ((graph)->mmaddrs) MapWrUns(buf, cnt, &((graph)->mmaddrs)); \
else WriteUns(buf, cnt, (graph)->addrs)

#define StkWrUns(buf,cnt,graph) \
if ((graph)->mmstack) MapWrUns(buf, cnt, &((graph)->mmstack)); \
else WriteUns(buf, cnt, (graph)->stack)

#define NdsReData(buf,cnt,graph) \
if ((graph)->mmnodes) MapReData(buf, cnt, (graph)->mmnodes); \
else ReadData(buf, cnt, (graph)->nodes)

#define AwsReData(buf,cnt,graph) \
if ((graph)->mmarrows) MapReData(buf, cnt, (graph)->mmarrows); \
else ReadData(buf, cnt, (graph)->arrows)

#define AdrReData(buf,cnt,graph) \
if ((graph)->mmaddrs) MapReData(buf, cnt, (graph)->mmaddrs); \
else ReadData(buf, cnt, (graph)->addrs)

#define StkReData(buf,cnt,graph) \
if ((graph)->mmstack) MapReData(buf, cnt, (graph)->mmstack); \
else ReadData(buf, cnt, (graph)->stack)

#define NdsWrData(buf,cnt,graph) \
if ((graph)->mmnodes) MapWrData(buf, cnt, &((graph)->mmnodes)); \
else WriteData(buf, cnt, (graph)->nodes)

#define AwsWrData(buf,cnt,graph) \
if ((graph)->mmarrows) MapWrData(buf, cnt, &((graph)->mmarrows)); \
else WriteData(buf, cnt, (graph)->arrows)

#define AdrWrData(buf,cnt,graph) \
if ((graph)->mmaddrs) MapWrData(buf, cnt, &((graph)->mmaddrs)); \
else WriteData(buf, cnt, (graph)->addrs)

#define StkWrData(buf,cnt,graph) \
if ((graph)->mmstack) MapWrData(buf, cnt, &((graph)->mmstack)); \
else WriteData(buf, cnt, (graph)->stack)

#define NdsReChr(buf,cnt,graph) \
if ((graph)->mmnodes) MapReChr(buf, cnt, (graph)->mmnodes); \
else ReadChr(buf, cnt, (graph)->nodes)

#define AwsReChr(buf,cnt,graph) \
if ((graph)->mmarrows) MapReChr(buf, cnt, (graph)->mmarrows); \
else ReadChr(buf, cnt, (graph)->arrows)

#define AdrReChr(buf,cnt,graph) \
if ((graph)->mmaddrs) MapReChr(buf, cnt, (graph)->mmaddrs); \
else ReadChr(buf, cnt, (graph)->addrs)

#define StkReChr(buf,cnt,graph) \
if ((graph)->mmstack) MapReChr(buf, cnt, (graph)->mmstack); \
else ReadChr(buf, cnt, (graph)->stack)

#define NdsWrChr(buf,cnt,graph) \
if ((graph)->mmnodes) MapWrChr(buf, cnt, &((graph)->mmnodes)); \
else WriteChr(buf, cnt, (graph)->nodes)

#define AwsWrChr(buf,cnt,graph) \
if ((graph)->mmarrows) MapWrChr(buf, cnt, &((graph)->mmarrows)); \
else WriteChr(buf, cnt, (graph)->arrows)

#define AdrWrChr(buf,cnt,graph) \
if ((graph)->mmaddrs) MapWrChr(buf, cnt, &((graph)->mmaddrs)); \
else WriteChr(buf, cnt, (graph)->addrs)

#define StkWrChr(buf,cnt,graph) \
if ((graph)->mmstack) MapWrChr(buf, cnt, &((graph)->mmstack)); \
else WriteChr(buf, cnt, (graph)->stack)

DE1(HASH *, HashCreate,
    unsigned long, size)
DE2(void, HashClear,
    HASH *, hash,
    unsigned long, size)
DE3(void, HashRead,
    HASH *, hash,
    unsigned long, size,
    BINARYFILE, file)
DE3(void, HashWrite,
    HASH *, hash,
    unsigned long, size,
    BINARYFILE, file)

/***************************************************************************
*
* Macro:     HashGet
* Purpose:   This macro tells us a value in an data position
*            in a segmented hash table.
* Arguments: hash - the table
*            number - the position
* Returns:   the value
* See also:  HashCreate, HashClear, HashPut, HashDiscard, HashRead,
*            HashWrite
*
****************************************************************************
*/
#define HashGet(hash, number) \
    ((unsigned long *) PtrBlGet(hash, (number) >> HLEAFBITS))\
    [(unsigned) ((number) & HLEAFMASK)]
    
/***************************************************************************
*
* Macro:     HashPut
* Purpose:   This macro sets a value in an data position
*            in a segmented hash table.
* Arguments: hash - the table
*            number - the position
*            value - the new value
* Returns:
* See also:  HashCreate, HashClear, HashGet, HashDiscard, HashRead,
*            HashWrite
*
****************************************************************************
*/
#define HashPut(hash, number, value) \
    (((unsigned long *) PtrBlGet(hash, (number) >> HLEAFBITS))\
     [(unsigned) ((number) & HLEAFMASK)] \
     = (value))

/***************************************************************************
*
* Macro:     HashDiscard
* Purpose:   This macro removes a segmented hash table from the memory.
* Arguments: hash - the table
* Returns:
* See also:  HashCreate, HashClear, HashGet, HashPut, HashRead, HashWrite
*
****************************************************************************
*/
#define HashDiscard(hash) PtrBlDestroy(hash, TRUE)
