/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  mark.h
*
* Part of: PROD's programs araprod, "gphgen", probe, and prpp
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See mark.c.
*
* Version history:
*     $Log$
***************************************************************************/

/* The TUPLE type is used to present a tuple in a place in a net. */
typedef struct TUPLE
{
    unsigned _arity;
    DOMAIN *_array;
} TUPLE;

/* The MULTUP type is used to present all copies of a tuple in a place
 * in a net. The elements of MULTUP are called multituples. */
typedef struct MULTUP
{
    unsigned long _count;
    unsigned long _freeCount;
    TUPLE *_tuple;
} MULTUP;

/* The MARKING type is used to present a marking of a place in a net. */
#define MARKING LIST

/* The STATE type is used to present a marking of a net. */
typedef MARKING *STATE;

extern unsigned _magNofPlaces, *_magProfile;
#define MagGNofPlaces() _magNofPlaces
#define MagGProfile(place) (_magProfile[place])
#define MagIProfile(place,value) (_magProfile[place] = (value))

#define StaGMarking(state,place) ((state)[place])
#define StaIMarking(state,place,marking) ((state)[place] = (marking))
#define StaAlloc() \
(MagGNofPlaces() ? \
 (STATE *) AllocArray(MagGNofPlaces(), sizeof(MARKING)) : \
 (STATE *) NULL)

#define MarEmpty(marking)  LstEmpty(marking)
#define MarBegin(marking)  ((MULTUP *) LstBegin(marking))
#define MarStep(marking)  ((MULTUP *) LstStep(marking))
#define MarThis(marking)  ((MULTUP *) LstThis(marking))
#define MarFirst(marking) ((MULTUP *) LstFirst(marking))
#define MarLast(marking) ((MULTUP *) LstLast(marking))
#define MarGet(marking)  LstGet(marking)
#define MarSet(marking,cursor)  LstSet(marking, cursor)
#define MarInsLast(marking,multup) \
    ((void) LstInsLast(marking, (char *) (multup)))
#define MarInsNext(marking,multup) \
    ((void) LstInsNext(marking, (char *) (multup)))
#define MarDelNext(marking)  ((void) LstDelNext(marking, TRUE))
#define MarCreate() LstCreate(&multupTools)
#define MarCopy(marking) LstCopy(marking, TRUE)
#define MarDiscard(marking) LstDestroy(marking, TRUE)

#define MARFOR(marking,multup) LSTFOR(marking, multup, MULTUP)

#define MultupAlloc() (MULTUP *) Allocate(sizeof(MULTUP))
#define MultupGCount(multup) ((multup)->_count)
#define MultupSCount(multup,count) ((multup)->_count = (count))
#define MultupICount(multup,count) ((multup)->_count += (count))
#define MultupDCount(multup,count) ((multup)->_count -= (count))
#define MultupGFreeCount(multup) ((multup)->_freeCount)
#define MultupSFreeCount(multup,freeCount) ((multup)->_freeCount = (freeCount))
#define MultupIFreeCount(multup,freeCount) ((multup)->_freeCount += (freeCount))
#define MultupDFreeCount(multup,freeCount) ((multup)->_freeCount -= (freeCount))
#define MultupGTuple(multup) ((multup)->_tuple)
#define MultupSTuple(multup,tuple) ((multup)->_tuple = (tuple))
#define MultupATuple(multup) (&((multup)->_tuple))
#define MultupGArity(multup) TupGArity((multup)->_tuple)
#define MultupGField(multup,field) TupGField((multup)->_tuple, field)
#define MultupAField(multup,field) TupAField((multup)->_tuple, field)
#define MultupSField(multup,field,value) TupSField((multup)->_tuple, field, value)
extern TOOLS multupTools;

#define TupAlloc() ((TUPLE *) Allocate(sizeof(TUPLE)))
#define TupGArity(tuple) ((tuple)->_arity)
#define TupAArity(tuple) (&((tuple)->_arity))
#define TupSArity(tuple,arity) ((tuple)->_arity = (arity))
#define TupGArray(tuple) ((tuple)->_array)
#define TupAArray(tuple) (&((tuple)->_array))
#define TupSArray(tuple,array) ((tuple)->_array = (array))
#define TupGField(tuple,field) ((tuple)->_array[field])
#define TupAField(tuple,field) ((tuple)->_array + (unsigned) (field))
#define TupSField(tuple,field,value) ((tuple)->_array[field] = (value))
#define TupIncField(tuple,field) ((tuple)->_array[field]++)
extern TOOLS tupTools;

DE1(STATE *, StaFrCopy,
    STATE *, state)
DE1(void, StaDiscard,
    STATE *, state)
#ifdef DEBUG
DE2(void, StaDump,
    STATE *, state,
    unsigned, depth)
DE2(void, MarDump,
    MARKING *, marking,
    unsigned, depth)
#endif
DE2(void, MarDelete,
    MARKING *, marking,
    MULTUP *, multup)
DE2(void, MarInsert,
    MARKING *, marking,
    MULTUP *, multup)
DE2(BOOL, MarLess,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(BOOL, MarLequal,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(BOOL, MarEqual,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(MARKING *, MarProduct,
    MARKING *, marking,
    unsigned long, coeff)
DE2(MARKING *, MarCartes,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(MARKING *, MarAdd,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(MARKING *, MarSubtract,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(MARKING *, MarIntersect,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(MARKING *, MarCatenate,
    MARKING *, marking1,
    MARKING *, marking2)
DE2(MULTUP *, MultupCons,
    unsigned long, count,
    TUPLE *, tuple)
DE1(MULTUP *, MultupCpy,
    MULTUP *, multup)
DE1(void, MultupDiscard,
    MULTUP *, multup)
#ifdef DEBUG
DE2(void, MultupDump,
    MULTUP *, multup,
    unsigned, depth)
#endif
DE1(TUPLE *, TupCons,
    unsigned, arity)
DE1(TUPLE *, TupCopy,
    TUPLE *, tuple)
DE1(void, TupDiscard,
    TUPLE *, tuple)
DE2(void, TupDump,
    TUPLE *, tuple,
    unsigned, depth)
DE2(int, TupComp,
    TUPLE *, tuple1,
    TUPLE *, tuple2)
