/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptnet.h
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     See ptandor.c, ptelem,c, ptnet.c and ptvisit.c.
*
* Version history:
*     $Log$
***************************************************************************/

/* The elements of PLC are called low level places. A low level place
 * is a place in a place/transition net and corresponds to a pair
 * of a high level place (a place of a predicate/transition net) and
 * a tuple. */
typedef struct PLC
{
    DATATYPE *_key;        /* an encoding of the high level correspondent */
    unsigned _flag;
    unsigned long _count;  /* the number of tokens */
    union
	{
	    LIST *_pureIn; /* arcs from {t | W(t,p) > 0 = W(p,t)} */
	    LIST *_next;   /* if the place corresponds to a state of a
                            * Buchi automaton, then the place is isolated
                            * and next is a list of places corresponding
                            * to the successor states */
	} _lp;
    LIST *_pureOut; /* arcs to {t | W(t,p) = 0 < W(p,t)} */
    LIST *_neutral; /* arcs from and to {t | W(t,p) = W(p,t) > 0} */
    LIST *_mostIn;  /* arcs from and to {t | W(t,p) > W(p,t) > 0} */
    LIST *_mostOut; /* arcs from and to {t | 0 < W(t,p) < W(p,t)} */
    union
	{
	    LIST *_orPredAnd;   /* in the deletion algorithm of the
                                 * stubborn set method, a place is
                                 * an "and" vertex in an "and/or" graph,
                                 * and _orPredAnd is the set of predecessor
                                 * vertices, all of which are "or"
                                 * transitions */
	    unsigned long _ind; /* _ind indicates the number of the place
				 * but is used only during the computation
                                 * of place symmetries in the symmetry
                                 * method (it is thus possible to use
                                 * the symmetry method together with
                                 * the deletion algorithm of the stubborn
                                 * set method) */
#ifdef SMODELS
	    struct
		{
		    unsigned long _outMax;
		    VOID *_gamma;
		    BALTREE *_ineqs;
		} _sp;
#endif
	} _up;
} PLC;

#define PlcAlloc() ((PLC *) Allocate(sizeof(PLC)))
extern TOOLS plcTools;
extern BALTOOLS plcBalTools;

#ifdef SMODELS

typedef struct INEQ
{
    unsigned long _card;
    VOID *_ge;
    VOID *_lt;
} INEQ;

#define IneqAlloc() ((INEQ *) Allocate(sizeof(INEQ)))
extern TOOLS ineqTools;
extern BALTOOLS ineqBalTools;

#define IneqGCard(ineq) ((ineq)->_card)
#define IneqCCard(ineq) ((ineq)->_card = (unsigned long) 0L)
#define IneqSCard(ineq,card) ((ineq)->_card = (card))
#define IneqACard(ineq) (&((ineq)->_card))
#define IneqGGe(ineq) ((ineq)->_ge)
#define IneqCGe(ineq) ((ineq)->_ge = NULL)
#define IneqSGe(ineq,ge) ((ineq)->_ge = (ge))
#define IneqAGe(ineq) (&((ineq)->_ge))
#define IneqGLt(ineq) ((ineq)->_lt)
#define IneqCLt(ineq) ((ineq)->_lt = NULL)
#define IneqSLt(ineq,lt) ((ineq)->_lt = (lt))
#define IneqALt(ineq) (&((ineq)->_lt))

DE1(INEQ *, IneqCons,
    unsigned long, card)
DE1(INEQ *, IneqCopy,
    INEQ *, ineq)
DE1(void, IneqDiscard,
    INEQ *, ineq)
#ifdef DEBUG
DE2(void, IneqDump,
    INEQ *, ineq,
    unsigned, depth)
#endif
DE2(int, IneqKeyCmp,
    INEQ *, ineq,
    unsigned long *, key)

#endif

/* The elements of TRN are called low level transitions. A low level
 * transition is a transition in a place/transition net and corresponds
 * to an instance of a high level transition (a transition in a
 * predicate/transition net). */
typedef struct TRN
{
    DATATYPE *_key; /* an encoding of the high level correspondent */
    unsigned long _number; /* the order number of the low level transition
                            * (the numbering starts from 0) */
    unsigned _flag;
    unsigned _where;
    VOID *_cursor;
    struct TRN *_down;
    unsigned *_block;
    LIST *_pureIn;  /* arcs from {p | W(p,t) > 0 = W(t,p)} */
    LIST *_pureOut; /* arcs to {p | W(p,t) = 0 < W(t,p)} */
    LIST *_neutral; /* arcs from and to {p | W(p,t) = W(t,p) > 0} */
    LIST *_mostIn;  /* arcs from and to {p | W(p,t) > W(t,p) > 0} */
    LIST *_mostOut; /* arcs from and to {p | 0 < W(p,t) < W(t,p)} */
    union
	{
	    LIST *_higher; /* if there are more than one precedence class,
                            * then _higher is a list of those transitions
                            * that priorize this transition */
	    union
		{
		    struct
			{
			    unsigned long _val; /* in the incremental
                                                 * algorithm of the stubborn
                                                 * set method, _val is
                                                 * a number associated with
                                                 * a transition vertex in an
                                                 * implicit dependency graph */
                            unsigned long _dupVal; /* _dupVal is used for
                                                    * storing and restoring
                                                    * of _val */
                            unsigned long _min;
			} _n;
		    struct /* this struct is needed in the deletion
                            * algorithm of the stubborn set method */
			{
			    LIST *_plcPredAO;      /* place predecessors
                                                    * of the transition
                                                    * in an "and/or" graph */
			    LIST *_orStrongPredAO; /* BINOR predecessors via
                                                    * "strong" edges */  
			    LIST *_orWeakPredAO;   /* BINOR predecessors via
						    * "weak" edges */
			    unsigned long _nOrSucc; /* if the transition is
                                                     * an "or" vertex, then
						     * _nOrSucc is the number
                                                     * of successors */
			    unsigned long _dupNOrSucc; /* _dupNOrSucc is used
							* for storing and
							* restoring of
							* _nOrSucc */
                            struct BINOR *_binOr;
			} _t;
#ifdef SMODELS
		    struct
			{
			    VOID *_plain;
			    VOID *_gamma;
			    VOID *_kappa;
			    VOID *_strong;
			    VOID *_weak;
			    VOID *_active;
			    VOID *_toStack;
			} _a;
#endif
		} _vt;
	} _ut;
} TRN;

#define TrnAlloc() ((TRN *) Allocate(sizeof(TRN)))
extern TOOLS trnTools;
extern BALTOOLS trnBalTools;

/* An element of PLARC is an arc or a zero-effect self-loop
 * connected to a place. */
typedef struct PLARC
{
    TRN *_trn;
    unsigned long _weight;
} PLARC;

#define PlArcAlloc() ((PLARC *) Allocate(sizeof(PLARC)))
extern TOOLS plArcTools;
extern BALTOOLS plArcBalTools;

/* An element of PLDRC is a non-zero-effect self-loop connected to a place. */
typedef struct PLDRC
{
    TRN *_trn;
    unsigned long _in;  /* W(t,p) */
    unsigned long _out; /* W(p,t) */
} PLDRC;

#define PlDrcAlloc() ((PLDRC *) Allocate(sizeof(PLDRC)))
extern TOOLS plDrcTools;
extern BALTOOLS plDrcBalTools;

/* An element of TRARC is an arc or a zero-effect self-loop connected
 * to a transition. */
typedef struct TRARC
{
    PLC *_plc;
    unsigned long _weight;
#ifdef SMODELS
    VOID *_gamma;
#endif
} TRARC;

#define TrArcAlloc() ((TRARC *) Allocate(sizeof(TRARC)))
extern TOOLS trArcTools;
extern BALTOOLS trArcBalTools;

/* An element of PLDRC is a non-zero-effect self-loop connected
 * to a transition. */
typedef struct TRDRC
{
    PLC *_plc;
    unsigned long _in;  /* W(p,t) */
    unsigned long _out; /* W(t,p) */
#ifdef SMODELS
    VOID *_gamma;
#endif
} TRDRC;

#define TrDrcAlloc() ((TRDRC *) Allocate(sizeof(TRDRC)))
extern TOOLS trDrcTools;
extern BALTOOLS trDrcBalTools;

/* An element of BINOR is a vertex in an "and/or" graph, needed
 * in the deletion algorithm of the stubborn set method. A transition
 * may be associated with more than one of such vertices. A BINOR
 * vertex is neither an "and" nor an "or" vertex. It has two kinds
 * of outgoing edges, "strong" and "weak". An "and/or" graph of PROD
 * can be transformed into a pure "and/or" graph of the kind described
 * in the reports A26 and B12 of DSL(HUT). The transformation is
 * as follows: let b be a BINOR vertex. Split b into two "and" vertices,
 * x and y. Let x have only the "strong" and y only the "weak" outgoing
 * edges of b. Then create a new "or" vertex, z, and redirect the incoming
 * edges of b to be the incoming edges of z. Finally, create two new edges,
 * one from z to x and one from z to y. */
typedef struct BINOR
{
    unsigned _flag;
    TRN *_trn;
} BINOR;

#define BinOrAlloc() ((BINOR *) Allocate(sizeof(BINOR)))
#define BinOrDiscard(binOr) Deallocate((char *) (binOr))
extern TOOLS binOrTools;

#define PlcGKey(plc) ((plc)->_key)
#define PlcAKey(plc) (&((plc)->_key))
#define PlcSKey(plc,key) ((plc)->_key = (key))
#define PlcCKey(plc) ((plc)->_key = NULL)
#define PlcDKey(plc) (Deallocate((char *) ((plc)->_key)), (plc)->_key = NULL)
#define PlcGInd(plc) ((plc)->_up._ind)
#define PlcAInd(plc) (&((plc)->_up._ind))
#define PlcSInd(plc,ind) ((plc)->_up._ind = (ind))
#define PlcGCount(plc) ((plc)->_count)
#define PlcACount(plc) (&((plc)->_count))
#define PlcSCount(plc,count) ((plc)->_count = (count))
#define PlcICount(plc,count) ((plc)->_count += (count))
#define PlcDCount(plc,count) ((plc)->_count -= (count))
#define PlcGPurIn(plc) ((plc)->_lp._pureIn)
#define PlcZPurIn(plc,list) ((plc)->_lp._pureIn = (list))
#define PlcSPurIn(plc) ((plc)->_lp._pureIn = LstCreate(&plArcTools))
#define PlcCPurIn(plc) ((plc)->_lp._pureIn = NULL)
#define PlcDPurIn(plc) \
    (LstDestroy((plc)->_lp._pureIn, TRUE), (plc)->_lp._pureIn = NULL)
#define PlcGNext(plc) ((plc)->_lp._next)
#define PlcZNext(plc,list) ((plc)->_lp._next = (list))
#define PlcSNext(plc) ((plc)->_lp._next = LstCreate(&plcTools))
#define PlcCNext(plc) ((plc)->_lp._next = NULL)
#define PlcDNext(plc) \
    (LstDestroy((plc)->_lp._next, TRUE), (plc)->_lp._next = NULL)
#define PlcGPurOut(plc) ((plc)->_pureOut)
#define PlcZPurOut(plc,list) ((plc)->_pureOut = (list))
#define PlcSPurOut(plc) ((plc)->_pureOut = LstCreate(&plArcTools))
#define PlcCPurOut(plc) ((plc)->_pureOut = NULL)
#define PlcDPurOut(plc) \
    (LstDestroy((plc)->_pureOut, TRUE), (plc)->_pureOut = NULL)
#define PlcGNeutral(plc) ((plc)->_neutral)
#define PlcZNeutral(plc,list) ((plc)->_neutral = (list))
#define PlcSNeutral(plc) ((plc)->_neutral = LstCreate(&plArcTools))
#define PlcCNeutral(plc) ((plc)->_neutral = NULL)
#define PlcGMosIn(plc) ((plc)->_mostIn)
#define PlcZMosIn(plc,list) ((plc)->_mostIn = (list))
#define PlcSMosIn(plc) ((plc)->_mostIn = LstCreate(&plDrcTools))
#define PlcCMosIn(plc) ((plc)->_mostIn = NULL)
#define PlcGMosOut(plc) ((plc)->_mostOut)
#define PlcZMosOut(plc,list) ((plc)->_mostOut = (list))
#define PlcSMosOut(plc) ((plc)->_mostOut = LstCreate(&plDrcTools))
#define PlcCMosOut(plc) ((plc)->_mostOut = NULL)
#define PlcGOrPredAnd(plc) ((plc)->_up._orPredAnd)
#define PlcZOrPredAnd(plc,list) ((plc)->_up._orPredAnd = (list))
#define PlcSOrPredAnd(plc) ((plc)->_up._orPredAnd = LstCreate(&trnTools))
#define PlcCOrPredAnd(plc) ((plc)->_up._orPredAnd = NULL)
#define PlcDOrPredAnd(plc) \
    (LstDestroy((plc)->_up._orPredAnd, FALSE), (plc)->_up._orPredAnd = NULL)
#ifdef SMODELS
#define PlcGOutMax(plc) ((plc)->_up._sp._outMax)
#define PlcSOutMax(plc,outMax) ((plc)->_up._sp._outMax = (outMax))
#define PlcCOutMax(plc) ((plc)->_up._sp._outMax = (unsigned long) 0L)
#define PlcGGamma(plc) ((plc)->_up._sp._gamma)
#define PlcSGamma(plc,gamma) ((plc)->_up._sp._gamma = (gamma))
#define PlcCGamma(plc) ((plc)->_up._sp._gamma = NULL)
#define PlcGIneqs(plc) ((plc)->_up._sp._ineqs)
#define PlcZIneqs(plc,tree) ((plc)->_up._sp._ineqs = (tree))
#define PlcSIneqs(plc) ((plc)->_up._sp._ineqs = BalTrCreate(&ineqBalTools))
#define PlcCIneqs(plc) ((plc)->_up._sp._ineqs = NULL)
#define PlcDIneqs(plc) \
    (BalTrDestroy((plc)->_up._sp._ineqs, TRUE), (plc)->_up._sp._ineqs = NULL)
#endif
#define PlcGFlag(plc) ((plc)->_flag)
#define PlcSFlag(plc,flag) ((plc)->_flag = (flag))
#define PlcCFlag(plc) ((plc)->_flag = 0)
#define PlcSatisfied(plc) ((plc)->_flag & 1)
#define PlcFeed(plc) ((plc)->_flag |= 1)
#define PlcThirsty(plc) ((plc)->_flag & 2)
#define PlcSalt(plc) ((plc)->_flag |= 2)
#define PlcWater(plc) ((plc)->_flag &= ~2)
#define PlcReset(plc) ((plc)->_flag &= ~3)
#define PlcReject(plc) ((plc)->_flag & 4)
#define PlcSReject(plc) ((plc)->_flag |= 4)
#define PlcDeadlock(plc) ((plc)->_flag & 8)
#define PlcSDeadlock(plc) ((plc)->_flag |= 8)
#define PlcLivelock(plc) ((plc)->_flag & 16)
#define PlcSLivelock(plc) ((plc)->_flag |= 16)
#define PlcInfinite(plc) ((plc)->_flag & 32)
#define PlcSInfinite(plc) ((plc)->_flag |= 32)
#define PlcBuchi(plc) ((plc)->_flag & 64)
#define PlcSBuchi(plc) ((plc)->_flag |= 64)
#define PlcMonitor(plc) ((plc)->_flag & 124)

#define TrnGKey(trn) ((trn)->_key)
#define TrnAKey(trn) (&((trn)->_key))
#define TrnSKey(trn,key) ((trn)->_key = (key))
#define TrnCKey(trn) ((trn)->_key = NULL)
#define TrnDKey(trn) (Deallocate((char *) ((trn)->_key)), (trn)->_key = NULL)
#define TrnGNumber(trn) ((trn)->_number)
#define TrnANumber(trn) (&((trn)->_number))
#define TrnSNumber(trn,number) ((trn)->_number = (number))
#define TrnCNumber(trn) ((trn)->_number = 0)
#define TrnGFlag(trn) ((trn)->_flag)
#define TrnSFlag(trn,flag) ((trn)->_flag = (flag))
#define TrnCFlag(trn) ((trn)->_flag = 0)
#define TrnEnabled(trn) ((trn)->_flag & 1)
#define TrnAccept(trn) ((trn)->_flag |= 1)
#define TrnSatisfied(trn) ((trn)->_flag & 2)
#define TrnFeed(trn) ((trn)->_flag |= 2)
#define TrnImpoverish(trn) ((trn)->_flag &= ~2)
#define TrnThirsty(trn) ((trn)->_flag & 4)
#define TrnSalt(trn) ((trn)->_flag |= 4)
#define TrnWater(trn) ((trn)->_flag &= ~4)
#define TrnAsleep(trn) ((trn)->_flag & 8)
#define TrnPutToSleep(trn) ((trn)->_flag |= 8)
#define TrnWake(trn) ((trn)->_flag &= ~8)
#define TrnOld(trn) ((trn)->_flag & 16)
#define TrnEnter(trn) ((trn)->_flag |= 16)
#define TrnLeave(trn) ((trn)->_flag &= ~16)
#define TrnReset(trn) ((trn)->_flag &= ~31)
#define TrnVisible(trn) ((trn)->_flag & 32)
#define TrnSVisible(trn) (_nTrVisible++, (trn)->_flag |= 32)
#define TrnInitial(trn) ((trn)->_flag & 64)
#define TrnSInitial(trn) ((trn)->_flag |= 64)
#define TrnDummy(trn) ((trn)->_flag & 128)
#define TrnSDummy(trn) ((trn)->_flag |= 128)
#define TrnFlagStore(trn) \
    ((trn)->_flag = ((trn)->_flag & 255) | (((trn)->_flag & 31) << 8))
#define TrnFlagRestore(trn) \
    ((trn)->_flag = ((trn)->_flag & 224) | (((trn)->_flag & 7936) >> 8))
#define TrnGWhere(trn) ((trn)->_where)
#define TrnSWhere(trn,where) ((trn)->_where = (where))
#define TrnGCursor(trn) ((trn)->_cursor)
#define TrnSCursor(trn,cursor) ((trn)->_cursor = (cursor))
#define TrnCCursor(trn) ((trn)->_cursor = NULL)
#define TrnGDown(trn) ((trn)->_down)
#define TrnSDown(trn,down) ((trn)->_down = (down))
#define TrnCDown(trn) ((trn)->_down = NULL)
#define TrnBlocks(trn) ((trn)->_block)
#define TrnCBlocks(trn) ((trn)->_block = NULL)
#define TrnABlocks(trn,n) \
((trn)->_block = (unsigned *) AllocArray(n, sizeof(unsigned)))
#define TrnSBlock(trn,i,v) ((trn)->_block[i] = (v))
#define TrnGBlock(trn,i) ((trn)->_block[i])
#define TrnDBlocks(trn) \
(Deallocate((char *) (trn)->_block), (trn)->_block = NULL)
#define TrnGPurIn(trn) ((trn)->_pureIn)
#define TrnZPurIn(trn,list) ((trn)->_pureIn = (list))
#define TrnSPurIn(trn) ((trn)->_pureIn = LstCreate(&trArcTools))
#define TrnCPurIn(trn) ((trn)->_pureIn = NULL)
#define TrnDPurIn(trn) \
    (LstDestroy((trn)->_pureIn, TRUE), (trn)->_pureIn = NULL)
#define TrnGPurOut(trn) ((trn)->_pureOut)
#define TrnZPurOut(trn,list) ((trn)->_pureOut = (list))
#define TrnSPurOut(trn) ((trn)->_pureOut = LstCreate(&trArcTools))
#define TrnCPurOut(trn) ((trn)->_pureOut = NULL)
#define TrnDPurOut(trn) \
    (LstDestroy((trn)->_pureOut, TRUE), (trn)->_pureOut = NULL)
#define TrnGNeutral(trn) ((trn)->_neutral)
#define TrnZNeutral(trn,list) ((trn)->_neutral = (list))
#define TrnSNeutral(trn) ((trn)->_neutral = LstCreate(&trArcTools))
#define TrnCNeutral(trn) ((trn)->_neutral = NULL)
#define TrnGMosIn(trn) ((trn)->_mostIn)
#define TrnZMosIn(trn,list) ((trn)->_mostIn = (list))
#define TrnSMosIn(trn) ((trn)->_mostIn = LstCreate(&trDrcTools))
#define TrnCMosIn(trn) ((trn)->_mostIn = NULL)
#define TrnGMosOut(trn) ((trn)->_mostOut)
#define TrnZMosOut(trn,list) ((trn)->_mostOut = (list))
#define TrnSMosOut(trn) ((trn)->_mostOut = LstCreate(&trDrcTools))
#define TrnCMosOut(trn) ((trn)->_mostOut = NULL)
#define TrnGHigher(trn) ((trn)->_ut._higher)
#define TrnZHigher(trn,list) ((trn)->_ut._higher = (list))
#define TrnSHigher(trn) ((trn)->_ut._higher = LstCreate(&trnTools))
#define TrnCHigher(trn) ((trn)->_ut._higher = NULL)
#define TrnGVal(trn) ((trn)->_ut._vt._n._val)
#define TrnSVal(trn,val) ((trn)->_ut._vt._n._val = (val))
#define TrnGDupVal(trn) ((trn)->_ut._vt._n._dupVal)
#define TrnSDupVal(trn,val) ((trn)->_ut._vt._n._dupVal = (val))
#define TrnGMin(trn) ((trn)->_ut._vt._n._min)
#define TrnSMin(trn,min) ((trn)->_ut._vt._n._min = (min))
#define TrnGNOrSucc(trn) ((trn)->_ut._vt._t._nOrSucc)
#define TrnSNOrSucc(trn,n) ((trn)->_ut._vt._t._nOrSucc = (n))
#define TrnCNOrSucc(trn) (((trn)->_ut._vt._t._nOrSucc) = 0L)
#define TrnINOrSucc(trn) (((trn)->_ut._vt._t._nOrSucc)++)
#define TrnDNOrSucc(trn) (((trn)->_ut._vt._t._nOrSucc)--)
#define TrnGDupNOrSucc(trn) ((trn)->_ut._vt._t._dupNOrSucc)
#define TrnSDupNOrSucc(trn,n) ((trn)->_ut._vt._t._dupNOrSucc = (n))
#define TrnCDupNOrSucc(trn) (((trn)->_ut._vt._t._dupNOrSucc) = 0L)
#define TrnGBinOr(trn) ((trn)->_ut._vt._t._binOr)
#define TrnSBinOr(trn,binOr) ((trn)->_ut._vt._t._binOr = (binOr))
#define TrnCBinOr(trn) ((trn)->_ut._vt._t._binOr = NULL)
#define TrnGPlcPredAO(trn) ((trn)->_ut._vt._t._plcPredAO)
#define TrnZPlcPredAO(trn,list) ((trn)->_ut._vt._t._plcPredAO = (list))
#define TrnSPlcPredAO(trn) ((trn)->_ut._vt._t._plcPredAO = LstCreate(&plcTools))
#define TrnCPlcPredAO(trn) ((trn)->_ut._vt._t._plcPredAO = NULL)
#define TrnDPlcPredAO(trn) \
    (LstDestroy((trn)->_ut._vt._t._plcPredAO, FALSE), \
     (trn)->_ut._vt._t._plcPredAO = NULL)
#define TrnGOrStrongPredAO(trn) ((trn)->_ut._vt._t._orStrongPredAO)
#define TrnZOrStrongPredAO(trn,list) ((trn)->_ut._vt._t._orStrongPredAO = (list))
#define TrnSOrStrongPredAO(trn) \
    ((trn)->_ut._vt._t._orStrongPredAO = LstCreate(&binOrTools))
#define TrnCOrStrongPredAO(trn) ((trn)->_ut._vt._t._orStrongPredAO = NULL)
#define TrnDOrStrongPredAO(trn) \
    (LstDestroy((trn)->_ut._vt._t._orStrongPredAO, FALSE), \
     (trn)->_ut._vt._t._orStrongPredAO = NULL)
#define TrnGOrWeakPredAO(trn) ((trn)->_ut._vt._t._orWeakPredAO)
#define TrnZOrWeakPredAO(trn,list) ((trn)->_ut._vt._t._orWeakPredAO = (list))
#define TrnSOrWeakPredAO(trn) \
    ((trn)->_ut._vt._t._orWeakPredAO = LstCreate(&binOrTools))
#define TrnCOrWeakPredAO(trn) ((trn)->_ut._vt._t._orWeakPredAO = NULL)
#define TrnDOrWeakPredAO(trn) \
    (LstDestroy((trn)->_ut._vt._t._orWeakPredAO, FALSE), \
     (trn)->_ut._vt._t._orWeakPredAO = NULL)
#ifdef SMODELS
#define TrnGPlain(trn) ((trn)->_ut._vt._a._plain)
#define TrnSPlain(trn,plain) ((trn)->_ut._vt._a._plain = (plain))
#define TrnCPlain(trn) ((trn)->_ut._vt._a._plain = NULL)
#define TrnGGamma(trn) ((trn)->_ut._vt._a._gamma)
#define TrnSGamma(trn,gamma) ((trn)->_ut._vt._a._gamma = (gamma))
#define TrnCGamma(trn) ((trn)->_ut._vt._a._gamma = NULL)
#define TrnGKappa(trn) ((trn)->_ut._vt._a._kappa)
#define TrnSKappa(trn,kappa) ((trn)->_ut._vt._a._kappa = (kappa))
#define TrnCKappa(trn) ((trn)->_ut._vt._a._kappa = NULL)
#define TrnGStrong(trn) ((trn)->_ut._vt._a._strong)
#define TrnSStrong(trn,strong) ((trn)->_ut._vt._a._strong = (strong))
#define TrnCStrong(trn) ((trn)->_ut._vt._a._strong = NULL)
#define TrnGWeak(trn) ((trn)->_ut._vt._a._weak)
#define TrnSWeak(trn,weak) ((trn)->_ut._vt._a._weak = (weak))
#define TrnCWeak(trn) ((trn)->_ut._vt._a._weak = NULL)
#define TrnGActive(trn) ((trn)->_ut._vt._a._active)
#define TrnSActive(trn,active) ((trn)->_ut._vt._a._active = (active))
#define TrnCActive(trn) ((trn)->_ut._vt._a._active = NULL)
#define TrnGToStack(trn) ((trn)->_ut._vt._a._toStack)
#define TrnSToStack(trn,toStack) ((trn)->_ut._vt._a._toStack = (toStack))
#define TrnCToStack(trn) ((trn)->_ut._vt._a._toStack = NULL)
#define TrnCAtoms(trn) \
    (TrnCPlain(trn), TrnCGamma(trn), TrnCKappa(trn), \
     TrnCStrong(trn), TrnCWeak(trn), TrnCActive(trn), TrnCToStack(trn))
#endif

#define PlArcGTrn(plArc) ((plArc)->_trn)
#define PlArcSTrn(plArc,trn) ((plArc)->_trn = (trn))
#define PlArcGWeight(plArc) ((plArc)->_weight)
#define PlArcAWeight(plArc) (&((plArc)->_weight))
#define PlArcSWeight(plArc,weight) ((plArc)->_weight = (weight))
#define PlArcIWeight(plArc,weight) ((plArc)->_weight += (weight))

#define PlDrcGTrn(plDrc) ((plDrc)->_trn)
#define PlDrcSTrn(plDrc,trn) ((plDrc)->_trn = (trn))
#define PlDrcGIn(plDrc) ((plDrc)->_in)
#define PlDrcAIn(plDrc) (&((plDrc)->_in))
#define PlDrcSIn(plDrc,in) ((plDrc)->_in = (in))
#define PlDrcGOut(plDrc) ((plDrc)->_out)
#define PlDrcAOut(plDrc) (&((plDrc)->_out))
#define PlDrcSOut(plDrc,out) ((plDrc)->_out = (out))

#define TrArcGPlc(trArc) ((trArc)->_plc)
#define TrArcSPlc(trArc,plc) ((trArc)->_plc = (plc))
#define TrArcGWeight(trArc) ((trArc)->_weight)
#define TrArcAWeight(trArc) (&((trArc)->_weight))
#define TrArcSWeight(trArc,weight) ((trArc)->_weight = (weight))
#define TrArcIWeight(trArc,weight) ((trArc)->_weight += (weight))
#ifdef SMODELS
#define TrArcGGamma(trArc) ((trArc)->_gamma)
#define TrArcSGamma(trArc,gamma) ((trArc)->_gamma = (gamma))
#define TrArcCGamma(trArc) ((trArc)->_gamma = NULL)
#endif

#define TrDrcGPlc(trDrc) ((trDrc)->_plc)
#define TrDrcSPlc(trDrc,plc) ((trDrc)->_plc = (plc))
#define TrDrcGIn(trDrc) ((trDrc)->_in)
#define TrDrcAIn(trDrc) (&((trDrc)->_in))
#define TrDrcSIn(trDrc,in) ((trDrc)->_in = (in))
#define TrDrcGOut(trDrc) ((trDrc)->_out)
#define TrDrcAOut(trDrc) (&((trDrc)->_out))
#define TrDrcSOut(trDrc,out) ((trDrc)->_out = (out))
#ifdef SMODELS
#define TrDrcGGamma(trDrc) ((trDrc)->_gamma)
#define TrDrcSGamma(trDrc,gamma) ((trDrc)->_gamma = (gamma))
#define TrDrcCGamma(trDrc) ((trDrc)->_gamma = NULL)
#endif

#define BinOrGFlag(binOr) ((binOr)->_flag)
#define BinOrSFlag(binOr,flag) ((binOr)->_flag = (flag))
#define BinOrCFlag(binOr) ((binOr)->_flag = 0)
#define BinOrGStrong(binOr) ((binOr)->_flag & 1)
#define BinOrSStrong(binOr) ((binOr)->_flag |= 1)
#define BinOrCStrong(binOr) ((binOr)->_flag &= ~1)
#define BinOrGWeak(binOr) ((binOr)->_flag & 2)
#define BinOrSWeak(binOr) ((binOr)->_flag |= 2)
#define BinOrCWeak(binOr) ((binOr)->_flag &= ~2)
#define BinOrHStrongNot(binOr) ((binOr)->_flag & 4)
#define BinOrTStrongNot(binOr) ((binOr)->_flag |= 4)
#define BinOrDStrongNot(binOr) ((binOr)->_flag &= ~4)
#define BinOrHWeakNot(binOr) ((binOr)->_flag & 8)
#define BinOrTWeakNot(binOr) ((binOr)->_flag |= 8)
#define BinOrDWeakNot(binOr) ((binOr)->_flag &= ~8)
#define BinOrGTrn(binOr) ((binOr)->_trn)
#define BinOrSTrn(binOr,trn) ((binOr)->_trn = (trn))

extern unsigned long _nTrVisible;
#define NTrVisible() _nTrVisible
extern PTRBL *_tblPlc;
#define TblPlc() _tblPlc
extern PTRBL *_dupTblPlc;
#define DupTblPlc() _dupTblPlc
extern PTRBL **_tblTrn;
#define TblTrn(class) _tblTrn[class]
extern unsigned *_higher;
#define Higher(class) (_higher[class] & 1)
#define SetHigher(class) (_higher[class] |= 1)
#define SelfHigher(class) (_higher[class] & 2)
#define SetSelfHigher(class) (_higher[class] |= 3)
#define Lower(class) (_higher[class] & 4)
#define SetLower(class) (_higher[class] |= 4)
#define Dependent(class) _higher[class]
extern BALTREE *_plcBalTree;
#define PlcBalTree() _plcBalTree
extern BALTREE **_trnBalTree;
#define TrnBalTree(class) _trnBalTree[class]
extern TRN *_lowTrn;
#define LowTrn() _lowTrn
#define SetLowTrn(trn) (_lowTrn = (trn))
extern unsigned _hiClass;
#define HiClass() _hiClass
#define SetHiClass(class) (_hiClass = (class))
extern BOOL _ptSy;
#define PtSymm() _ptSy
#define SetPtSymm() (_ptSy = TRUE)
#define ClPtSymm() (_ptSy = FALSE)
extern unsigned _ptVerbose;
#define PtVerbose() _ptVerbose
#define SetPtVerbose() (_ptVerbose |= 1)
#define PtSuperVerbose() (_ptVerbose & 2)
#define SetPtSuperVerbose() (_ptVerbose |= 2)
#define ClPtVerbose() (_ptVerbose = 0)
extern BOOL _ptStub;
#define PtStub() _ptStub
#define SetPtStub() (_ptStub = TRUE)
#define ClPtStub() (_ptStub = FALSE)
extern BOOL _ptCFFD;
#define PtCFFD() _ptCFFD
#define SetPtCFFD() (_ptCFFD = TRUE)
#define ClPtCFFD() (_ptCFFD = FALSE)
extern BOOL _ptDefVis;
#define PtDefVis() _ptDefVis
#define SetPtDefVis() (_ptDefVis = TRUE)
#define ClPtDefVis() (_ptDefVis = FALSE)
extern BOOL _ptSleep;
#define PtSleep() _ptSleep
#define SetPtSleep() (_ptSleep = TRUE)
#define ClPtSleep() (_ptSleep = FALSE)
extern BOOL _plRnd;
#define PlRnd() _plRnd
#define SetPlRnd() (_plRnd = TRUE)
#define ResetPlRnd() (_plRnd = FALSE)
extern BOOL _ptLazy;
#define PtLazy() _ptLazy
#define SetPtLazy() (_ptLazy = TRUE)
#define ClPtLazy() (_ptLazy = FALSE)
extern BOOL _trRnd;
#define TrRnd() _trRnd
#define SetTrRnd() (_trRnd = TRUE)
#define ResetTrRnd() (_trRnd = FALSE)
extern BOOL _ptDelAlg;
#define PtDelAlg() _ptDelAlg
#define SetPtDelAlg() (_ptDelAlg = TRUE)
#define ClPtDelAlg() (_ptDelAlg = FALSE)
extern BOOL _ptTryAgain;
#define PtTryAgain() _ptTryAgain
#define SetPtTryAgain() (_ptTryAgain = TRUE)
#define ClPtTryAgain() (_ptTryAgain = FALSE)
extern LIST *_binOrList;
#define BinOrList() _binOrList
#define BinOrSList() (_binOrList = LstCreate(&binOrTools))
#define BinOrDList() LstDestroy(_binOrList, TRUE)
extern LIST *_ptList;
#define PtList() _ptList
#define PtSList() (_ptList = LstCreate(&trnTools))
#define PtDList() LstDestroy(_ptList, FALSE)
extern unsigned long _ptSgCount;
#define PtSgCount() _ptSgCount
#define PtCSgCount() (_ptSgCount = 0L)
#define PtISgCount() (_ptSgCount++)
extern unsigned long _ptECnt;
#define PtGECnt() _ptECnt
#define PtCECnt() (_ptECnt = 0L)
#define PtIECnt() (_ptECnt++)
extern unsigned long _ptInvisCnt;
#define PtGInvisCnt() _ptInvisCnt
#define PtCInvisCnt() (_ptInvisCnt = 0L)
#define PtIInvisCnt() (_ptInvisCnt++)
extern unsigned long _ptNow;
#define PtNow() _ptNow
#define PtCNow() (_ptNow = 0L)
#define PtINow() (_ptNow++)
extern unsigned long _ptBest;
#define PtBest() _ptBest
#define PtSetBest(b) (_ptBest = (b))
extern unsigned long _ptSzBest;
#define PtSzBest() _ptSzBest
#define PtSzSBest(sz) (_ptSzBest = (sz))
extern unsigned long _ptCounter;
#define PtSCounter(c) (_ptCounter = (c))
#define PtDCounter() (_ptCounter--)
extern TRN *_ptKingTrn;
#define PtKingTrn() _ptKingTrn
#define PtSKingTrn(trn) (_ptKingTrn = (trn))
extern unsigned _ptClass;
#define PtClass() _ptClass
#define PtCClass() (_ptClass = 0)
#define PtIClass() (_ptClass++)
extern BOOL _ptRecurs;
#define PtRecurs() _ptRecurs
#define SetPtRecurs() (_ptRecurs = TRUE)
#define ClPtRecurs() (_ptRecurs = FALSE)
extern BALTREE *_ptSlpSet[4];
#define PtSlpSet() (_ptSlpSet[0])
#define PtASlpSet() (&(_ptSlpSet[0]))
#define PtSSlpSet(slpSet) (_ptSlpSet[0] = (slpSet))
#define PtDSlpSet() (BalTrDestroy(_ptSlpSet[0], TRUE), _ptSlpSet[0] = NULL)
#define Pt2SlpSet() (_ptSlpSet[1])
#define PtA2SlpSet() (&(_ptSlpSet[1]))
#define PtS2SlpSet(slpSet) (_ptSlpSet[1] = (slpSet))
#define PtD2SlpSet() (BalTrDestroy(_ptSlpSet[1], TRUE), _ptSlpSet[1] = NULL)
#define Pt3SlpSet() (_ptSlpSet[2])
#define PtA3SlpSet() (&(_ptSlpSet[2]))
#define PtS3SlpSet(slpSet) (_ptSlpSet[2] = (slpSet))
#define PtD3SlpSet() (BalTrDestroy(_ptSlpSet[2], TRUE), _ptSlpSet[2] = NULL)
#define Pt4SlpSet() (_ptSlpSet[3])
#define PtA4SlpSet() (&(_ptSlpSet[3]))
#define PtS4SlpSet(slpSet) (_ptSlpSet[3] = (slpSet))
#define PtD4SlpSet() (BalTrDestroy(_ptSlpSet[3], TRUE), _ptSlpSet[3] = NULL)
extern PLC *_ptFlyPlc;
#define PtFlyPlc() _ptFlyPlc
#define PtSFlyPlc(plc) (_ptFlyPlc = (plc))
extern BOOL ptDoIt;
extern BOOL ptBranching;
extern TRN *ptFirst;
extern PLC *buPlc;
extern FOOL _buchiOK;

DE0(void, FndEnbl)
DE0(void, PtGphUpdate)
DE2(void, PtC1,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE2(void, PtC2,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE2(void, PtC3,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE2(void, PtC4,
    unsigned long, awAddr,
    unsigned long, ndAddr)
DE0(void, PtOnTheFly)
DE0(void, PtInfinite)
DE0(void, PtIncrementalAlg)
DE0(void, PtDeletionAlg)
DE0(void, PtFollowers)
DE0(void, PtExpand)
DE1(BOOL, TrnChkEnbl,
    TRN *, trn)
DE1(BOOL, TrnAntiVisit,
    TRN *, trn)
DE1(BOOL, TrnStub,
    TRN *, trn)
DE1(void, TrnNaps,
    TRN *, trn)
DE1(void, TrnSpan,
    TRN *, trn)
DE1(BOOL, TrnVisStub,
    TRN *, trn)
DE1(void, TrnOA,
    TRN *, trn)
DE1(void, TrnAO,
    TRN *, trn)
DE1(BOOL, TrnAndOr,
    TRN *, trn)
DE0(PLC *, PlcCons)
DE1(PLC *, PlcCopy,
    PLC *, plc)
DE1(void, PlcDiscard,
    PLC *, plc)
#ifdef DEBUG
DE2(void, PlcDump,
    PLC *, plc,
    unsigned, depth)
#endif
DE2(int, PlcKeyCmp,
    PLC *, plc,
    DATATYPE *, key)
DE0(TRN *, TrnCons)
DE1(TRN *, TrnCopy,
    TRN *, trn)
DE1(void, TrnDiscard,
    TRN *, trn)
#ifdef DEBUG
DE2(void, TrnDump,
    TRN *, trn,
    unsigned, depth)
#endif
DE2(int, TrnKeyCmp,
    TRN *, trn,
    DATATYPE *, key)
DE2(PLARC *, PlArcCons,
    TRN *, trn,
    unsigned long, weight)
DE1(PLARC *, PlArcCopy,
    PLARC *, plArc)
#ifdef DEBUG
DE2(void, PlArcDump,
    PLARC *, plArc,
    unsigned, depth)
#endif
DE2(int, PlArcCompare,
    PLARC *, plArc1,
    PLARC *, plArc2)
DE3(PLDRC *, PlDrcCons,
    TRN *, trn,
    unsigned long, in,
    unsigned long, out)
DE1(PLDRC *, PlDrcCopy,
    PLDRC *, plDrc)
#ifdef DEBUG
DE2(void, PlDrcDump,
    PLDRC *, plDrc,
    unsigned, depth)
#endif
DE2(int, PlDrcCompare,
    PLDRC *, plDrc1,
    PLDRC *, plDrc2)
DE2(TRARC *, TrArcCons,
    PLC *, plc,
    unsigned long, weight)
DE1(TRARC *, TrArcCopy,
    TRARC *, trArc)
#ifdef DEBUG
DE2(void, TrArcDump,
    TRARC *, trArc,
    unsigned, depth)
#endif
DE2(int, TrArcCompare,
    TRARC *, trArc1,
    TRARC *, trArc2)
DE3(TRDRC *, TrDrcCons,
    PLC *, plc,
    unsigned long, in,
    unsigned long, out)
DE1(TRDRC *, TrDrcCopy,
    TRDRC *, trDrc)
#ifdef DEBUG
DE2(void, TrDrcDump,
    TRDRC *, trDrc,
    unsigned, depth)
#endif
DE2(int, TrDrcCompare,
    TRDRC *, trDrc1,
    TRDRC *, trDrc2)
DE1(BINOR *, BinOrCons,
    TRN *, trn)
DE1(BINOR *, BinOrCopy,
    BINOR *, binOr)
#ifdef DEBUG
DE2(void, BinOrDump,
    BINOR *, binOr,
    unsigned, depth)
#endif
DE4(void, PutInArc,
    TRN *, trn,
    unsigned long, weight,
    unsigned, plIndex,
    TUPLE *, tuple)
DE4(void, PutOutArc,
    TRN *, trn,
    unsigned long, weight,
    unsigned, plIndex,
    TUPLE *, tuple)
DE1(void, TrnReform,
    TRN *, trn)
DE0(void, PlcFill)
DE0(void, TrnFill)
DE1(void, PtOutput,
    BINARYFILE, file)
DE0(void, PtAraOutput)
DE1(void, PtPepOutput,
    char *, name)
DE1(BOOL, PlcIniFlag,
    PLC *, plc)
DE1(BOOL, PlcAOClean,
    PLC *, plc)
DE3(unsigned, PlcEncode,
    PTRBL *, tbl,
    DATATYPE **, buffer,
    unsigned *, size)
DE1(void, PlcDecode,
    DATATYPE *, buffer)
DE1(BOOL, TrnPrec,
    TRN *, trn)
DE1(void, PtInit,
    BINARYFILE, file)
DE0(void, PtLTSInit)
DE1(void, PtSRandom,
    UNS32, seedVal)
DE0(UNS32, PtTerm)
DE1(UNS32, PtRandom,
    UNS32, domSize)
DE2(void, LTSBegin,
    char *, name,
    BOOL, separate)
DE0(void, LTSEnd)
DE0(STATE *, PtState)
DE1(PLC *, GetPlc,
    DATATYPE *, key)
DE3(TRN *, GetTrn,
    unsigned, class,
    DATATYPE *, key,
    BOOL *, new)
DE1(void, TrnFire,
    TRN *, trn)
DE1(void, TrnLookStack,
    TRN *, trn)

extern TRN *ptInvFirst;
extern unsigned long ptM;
extern TRN *ptTrn;
extern TRN *pt2Trn;
extern PLC *ptPlc;
extern TRARC *ptTrArc;
extern TRDRC *ptTrDrc;
extern PLARC *ptPlArc;
extern PLDRC *ptPlDrc;
extern BINOR *ptBinOr;
extern TRN *ptVisible;
extern TRN *ptInvisible;
extern unsigned long ptSz;
extern unsigned long ptV;
extern unsigned long ptLimit;
extern unsigned long trNumber;

DE1(BOOL, TrnEnblCheck,
    TRN *, trn)
DE2(BOOL, AfterEnbl,
    TRN *, trn1,
    TRN *, trn2)
DE1(BOOL, TrnBest,
    TRN *, trn)
DE1(BOOL, TrnPrecCheck,
    TRN *, trn)
DE1(BOOL, TrnPrecCheck,
    TRN *, trn)
DE1(BOOL, TrnReCheck,
    TRN *, trn)
DE1(BOOL, PtNewSleep,
    unsigned long *, a)
DE1(BOOL, PtOldSleep,
    unsigned long *, a)
DE1(BOOL, TrnChkRe,
    TRN *, trn)
DE1(BOOL, TrnCntCand,
    TRN *, trn)
DE1(BOOL, TrnCntICand,
    TRN *, trn)
DE1(BOOL, TrnVRemoved,
    TRN *, trn)
DE1(BOOL, TrnWorseVis,
    TRN *, trn)
DE1(BOOL, TrnFndStrong,
    TRN *, trn)
DE1(void, TrnYrd,
    TRN *, trn)
DE1(void, TrnDry,
    TRN *, trn)
DE1(BOOL, TrnTryDel,
    TRN *, trn)
DE1(void, TrnLecnac,
    TRN *, trn)
DE1(void, TrnCancel,
    TRN *, trn)
DE1(void, TrnTewTas,
    TRN *, trn)
DE1(void, TrnSatWet,
    TRN *, trn)
DE1(BOOL, TrnPreAO,
    TRN *, trn)
DE1(BOOL, TrnAOClean,
    TRN *, trn)
DE1(BOOL, TrnDelCheck,
    TRN *, trn)
DE1(BOOL, TrnRedInc,
    TRN *, trn)
DE1(BOOL, TrnRedDel,
    TRN *, trn)
DE0(void, PtInside)
DE0(void, PtActual)
DE0(void, PtInner)
DE0(void, PtBuchi)
DE0(void, PtProCFFD)
DE0(void, PtRedInc)
DE1(BOOL, TrnDupOr,
    TRN *, trn)
DE1(BOOL, TrnBackup,
    TRN *, trn)
DE1(BOOL, TrnRehab,
    TRN *, trn)
DE1(BOOL, TrnRemember,
    TRN *, trn)
DE1(BOOL, TrnCollect,
    TRN *, trn)
DE0(void, Collect)
DE1(BOOL, TrnSCollect,
    TRN *, trn)
DE0(void, SCollect)
DE1(BOOL, TrnCntV,
    TRN *, trn)
DE1(BOOL, TrnVCollect,
    TRN *, trn)
DE0(void, VCollect)
DE0(void, WCollect)
DE6(void, DelThem,
    unsigned long, nInitial,
    unsigned long, limCand,
    BOOL, vremCheck,
    BOOL, fromScratch,
    BOOL, tp_visible,
    BOOL, sCollected)
DE0(void, Upsilon)
DE1(void, PtRedDel,
    BOOL, sCollected)
DE1(BOOL, TrnDsPrIn,
    TRN *, trn)
DE1(BOOL, TrnDsNeutral,
    TRN *, trn)
DE1(BOOL, TrnDsMsIn,
    TRN *, trn)
DE1(BOOL, TrnDsMsOut,
    TRN *, trn)
DE1(void, TrnEnNeutral,
    TRN *, trn)
DE1(void, TrnEnMsOut,
    TRN *, trn)
DE1(void, TrnEnPrIn,
    TRN *, trn)
DE1(void, TrnEnMsIn,
    TRN *, trn)
DE1(void, TrnTivis,
    TRN *, trn)
DE1(void, TrnVisit,
    TRN *, trn)
DE1(BOOL, TrnDvPrIn,
    TRN *, trn)
DE1(BOOL, TrnDvNeutral,
    TRN *, trn)
DE1(BOOL, TrnDvMsIn,
    TRN *, trn)
DE1(BOOL, TrnDvMsOut,
    TRN *, trn)
DE1(void, TrnEvNeutral,
    TRN *, trn)
DE1(void, TrnEvMsOut,
    TRN *, trn)
DE1(void, TrnEvPrIn,
    TRN *, trn)
DE1(void, TrnEvMsIn,
    TRN *, trn)
DE1(void, TrnNdPrIn,
    TRN *, trn)
DE1(void, TrnNdMsIn,
    TRN *, trn)
DE1(void, TrnOrPrIn,
    TRN *, trn)
DE1(void, TrnOrNeutral,
    TRN *, trn)
DE1(void, TrnOrMsIn,
    TRN *, trn)
DE1(void, TrnOrMsOut,
    TRN *, trn)

extern LIST **_visBlock;
extern char *_visColour;
#define ABlocks(n) (_visBlock = (LIST **) AllocArray(n, sizeof(LIST *)), \
_visColour = (char *) Allocate(n))
#define SBlock(i) (_visBlock[i] = LstCreate(&trnTools), _visColour[i] = 'y')
#define CBlock(i) (_visBlock[i] = NULL, _visColour[i] = '\0')
#define Block(i) (_visBlock[i])
#define Yellow(i) ('y' == _visColour[i])
#define Orange(i) ('o' == _visColour[i])
#define Blue(i) ('b' == _visColour[i])
#define SetYellow(i) (_visColour[i] = 'y')
#define SetOrange(i) (_visColour[i] = 'o')
#define SetBlue(i) (_visColour[i] = 'b')

extern BOOL _blocked;
#define Blocked() _blocked
#define SetBlocked() (_blocked = TRUE)
#define ClBlocked() (_blocked = FALSE)

extern PTRBL *visibles;

#ifdef SMODELS
extern BOOL _ptSmodels;
#define PtSmodels() _ptSmodels
#define SetPtSmodels() (_ptSmodels = TRUE)
#define ClPtSmodels() (_ptSmodels = FALSE)
extern BOOL _ptLongPlay;
#define PtLongPlay() _ptLongPlay
#define SetPtLongPlay() (_ptLongPlay = TRUE)
#define ClPtLongPlay() (_ptLongPlay = FALSE)
extern INEQ *ptIneq;
DE0(void, PtDelegate)
DE1(void, MkPlcGamma,
    PLC *, plc)
DE2(INEQ *, MkPlcIneq,
    PLC *, plc,
    unsigned long, card)
DE1(void, MkTrnGamma,
    TRN *, trn)
DE1(void, MkTrnPlain,
    TRN *, trn)
DE1(void, MkTrnKappa,
    TRN *, trn)
DE1(void, MkTrnActive,
    TRN *, trn)
DE1(void, MkTrnToStack,
    TRN *, trn)
#ifdef DEBUG
DE2(void, MkTrArcGamma,
    TRN *, trn,
    TRARC *, trArc)
#define MkTAGamma(trn,trArc) MkTrArcGamma(trn, trArc)
DE2(void, MkTrDrcGamma,
    TRN *, trn,
    TRDRC *, trDrc)
#define MkTDGamma(trn,trDrc) MkTrDrcGamma(trn, trDrc)
DE2(void, MkTrPl,
    TRN *, trn,
    PLC *, plc)
#define MkTP(trn,plc) MkTrPl(trn, plc)
#else
DE1(void, MkTrArcGamma,
    TRARC *, trArc)
#define MkTAGamma(trn,trArc) MkTrArcGamma(trArc)
DE1(void, MkTrDrcGamma,
    TRDRC *, trDrc)
#define MkTDGamma(trn,trDrc) MkTrDrcGamma(trDrc)
DE1(void, MkTrPl,
    TRN *, trn)
#define MkTP(trn,plc) MkTrPl(trn)
#endif
DE2(void, SimpleRule,
    VOID *, head,
    VOID *, tail)
DE2(void, ReverseRules,
    VOID *, atom1,
    VOID *, atom2)
#endif
